# pylint: disable=line-too-long
# fmt: off
"""
A version module managed by the invoke-release task

See also tasks.py
"""

__version_info__ = (0, 13, 0)
__version__ = '-'.join(filter(None, ['.'.join(map(str, __version_info__[:3])), (__version_info__[3:] or [None])[0]]))
# fmt: on



"""
S3URI
-----
"""
import json
import re
from dataclasses import dataclass
from datetime import datetime
from pathlib import PurePosixPath
from typing import Dict
from typing import Iterable
from typing import Pattern
from typing import Union

import boto3
import botocore
import botocore.exceptions

from aio_aws.logger import get_logger

LOGGER = get_logger(__name__)

BUCKET_REGEX = re.compile(r"^[a-z0-9][a-z0-9.-]{2,62}$")  # type: Pattern[str]


def bucket_validate(bucket_name: str) -> bool:
    """
    For technical details on S3 buckets, refer to

    - https://docs.aws.amazon.com/AmazonS3/latest/dev/BucketRestrictions.html

    The following are the rules for naming S3 buckets in all AWS Regions:

    - Bucket names must follow DNS-compliant naming conventions. Amazon S3 no
      longer supports creating bucket names that contain uppercase letters or
      underscores. This ensures that each bucket can be addressed using virtual
      host style addressing, such as https://myawsbucket.s3.amazonaws.com.
    - Bucket names must be unique across all existing bucket names in Amazon S3.
    - Bucket names must comply with DNS naming conventions.
    - Bucket names must be at least 3 and no more than 63 characters long.
    - Bucket names must not contain uppercase characters or underscores.
    - Bucket names must start with a lowercase letter or number.
    - Bucket names must be a series of one or more labels. Adjacent labels are
      separated by a single period (.). Bucket names can contain lowercase letters,
      numbers, and hyphens. Each label must start and end with a lowercase letter
      or a number.
    - Bucket names must not be formatted as an IP address (e.g., 192.168.5.4).
    - When you use virtual hosted–style buckets with Secure Sockets Layer (SSL),
      the SSL wildcard certificate only matches buckets that don't contain periods.
      To work around this, use HTTP or write your own certificate verification
      logic. We recommend that you do not use periods (".") in bucket names when
      using virtual hosted-style buckets.

    :param bucket_name:
    :return: True if the bucket_name is valid
    :raises: ValueError if the bucket_name is not valid
    """
    if BUCKET_REGEX.match(bucket_name):
        return True
    raise ValueError(f"The bucket_name is invalid: {bucket_name}")


@dataclass(frozen=True)
class S3Object:
    """Just the bucket_name and key for an :code:`s3.ObjectSummary`.

    This simple dataclass should work around problems with
    :code:`Pickle` for an :code:`s3.ObjectSummary`.

    .. code-block:: python

        from aio_aws.s3_uri import S3Object

        # assume obj is an s3.ObjectSummary
        S3Object(bucket=obj.bucket_name, key=obj.key)

    """

    bucket: str
    key: str

    @property
    def s3_uri(self):
        """s3_uri: str for :code:`s3://{bucket}/{key}`"""
        path = PurePosixPath(self.bucket) / PurePosixPath(self.key)
        return f"s3://{path}"


class S3Paths:
    """
    S3 URI components
    """

    def __init__(self, s3_uri: str = "", bucket: str = "", key: str = ""):
        """
        This class is usually instantiated by py:meth:`S3Paths.parse_s3_uri`
        rather than directly instantiated.  If it is instantiated directly,
        it takes either an s3_uri (which takes precedence) or it takes
        both a bucket and a key.

        :param s3_uri: the URI in the form of :code:`s3://{bucket}/{key}`
        :param bucket: the bucket in :code:`s3://{bucket}/{key}`
        :param key: the key in :code:`s3://{bucket}/{key}`
        """

        if s3_uri:
            s3_paths = S3Paths.parse_s3_uri(s3_uri)
            self._bucket = s3_paths.bucket
            self._key = s3_paths.key
        elif bucket:
            if not S3Paths.is_valid_bucket(bucket):
                raise ValueError(f"The bucket is invalid: {bucket}")
            self._bucket = bucket
            self._key = key
        else:
            raise ValueError("Failed to provide s3_uri or a valid bucket and key")

        self._key_path = ""
        self._key_file = ""

        if self._key:
            key = PurePosixPath(self._key)

            if key.suffix:
                self._key_path = str(key.parent)
                self._key_file = key.stem + key.suffix
            else:
                self._key_path = str(key)

    @property
    def protocol(self) -> str:
        return "s3://"

    @property
    def bucket(self) -> str:
        return self._bucket

    @property
    def key(self) -> str:
        return self._key

    @property
    def key_path(self) -> str:
        return self._key_path

    @property
    def key_file(self) -> str:
        return self._key_file

    @property
    def s3_uri(self):
        """s3_uri: str for :code:`s3://{bucket}/{key}`"""
        path = PurePosixPath(self.bucket) / PurePosixPath(self.key)
        return f"s3://{path}"

    def as_uri(self, protocol: str = "s3://") -> str:
        """a URI for :code:`{protocol}{bucket}/{key}`"""
        path = PurePosixPath(self.bucket) / PurePosixPath(self.key)
        return f"{protocol}{path}"

    @staticmethod
    def is_valid_bucket(bucket: str) -> bool:
        if BUCKET_REGEX.match(bucket):
            return True
        return False

    @staticmethod
    def parse_s3_uri(s3_uri: str) -> "S3Paths":
        """Parse an S3 URI into components; the delimiter must be '/'

        :param s3_uri:
        :return: :code:`S3Paths(bucket, key, key_path, key_file)`
        """
        s3_uri = str(s3_uri).strip()
        if not s3_uri.startswith("s3://"):
            raise ValueError(f"The s3_uri is invalid: {s3_uri}")

        paths = s3_uri.replace("s3://", "").split("/")
        if not any(paths):
            raise ValueError(f"The s3_uri is invalid: {s3_uri}")

        bucket = paths.pop(0)
        if not BUCKET_REGEX.match(bucket):
            raise ValueError(f"The s3_uri is invalid: {s3_uri}")

        key = ""
        if paths:
            key = str(PurePosixPath(*paths))

        return S3Paths(bucket=bucket, key=key)

    def glob_pattern(self, glob_pattern: str = "**/*"):
        """
        Construct a glob pattern relative to the key_path.  The final
        glob pattern is: :code:`str(PurePosixPath(self.key_path) / glob_pattern)`

        :param glob_pattern: str defaults to :code:`'**/*'`
        :return: str
        """
        return str(PurePosixPath(self.key_path) / glob_pattern)

    def glob_file_pattern(self):
        """
        Construct a glob pattern relative to the key_path.  The final
        glob pattern is: :code:`{key_path}/**/{file_stem}*.*`

        :return: str
        """
        file_stem = PurePosixPath(self.key).stem
        return f"{self.key_path}/**/{file_stem}*.*"

    def __eq__(self, other):
        if not hasattr(other, "s3_uri"):
            return NotImplemented
        return self.s3_uri == other.s3_uri

    def __ne__(self, other):
        if not hasattr(other, "s3_uri"):
            return NotImplemented
        return self.s3_uri != other.s3_uri

    def __lt__(self, other):
        if not hasattr(other, "s3_uri"):
            return NotImplemented
        return self.s3_uri < other.s3_uri

    def __le__(self, other):
        if not hasattr(other, "s3_uri"):
            return NotImplemented
        return self.s3_uri <= other.s3_uri

    def __ge__(self, other):
        if not hasattr(other, "s3_uri"):
            return NotImplemented
        return self.s3_uri >= other.s3_uri

    def __gt__(self, other):
        if not hasattr(other, "s3_uri"):
            return NotImplemented
        return self.s3_uri > other.s3_uri

    def __hash__(self):
        return hash(self.s3_uri)

    def __repr__(self):
        return "%s:%r" % (self.__class__, self.s3_uri)

    def __str__(self):
        return self.s3_uri


class S3URI(S3Paths):
    """S3 URI representation and parsing.

    At present, this class is designed to represent one s3 object.
    Although it has some utilities for working with general bucket and key
    paths, it is designed to work with one s3 object.

    For technical details on S3 Objects, refer to

    - https://docs.aws.amazon.com/AmazonS3/latest/dev/BucketRestrictions.html
    - https://docs.aws.amazon.com/AmazonS3/latest/dev/UsingMetadata.html

    The following are the rules for naming S3 buckets in all AWS Regions:

    - Bucket names must follow DNS-compliant naming conventions. Amazon S3 no
      longer supports creating bucket names that contain uppercase letters or
      underscores. This ensures that each bucket can be addressed using virtual
      host style addressing, such as https://myawsbucket.s3.amazonaws.com.
    - Bucket names must be unique across all existing bucket names in Amazon S3.
    - Bucket names must comply with DNS naming conventions.
    - Bucket names must be at least 3 and no more than 63 characters long.
    - Bucket names must not contain uppercase characters or underscores.
    - Bucket names must start with a lowercase letter or number.
    - Bucket names must be a series of one or more labels. Adjacent labels are
      separated by a single period (.). Bucket names can contain lowercase letters,
      numbers, and hyphens. Each label must start and end with a lowercase letter
      or a number.
    - Bucket names must not be formatted as an IP address (e.g., 192.168.5.4).
    - When you use virtual hosted–style buckets with Secure Sockets Layer (SSL),
      the SSL wildcard certificate only matches buckets that don't contain periods.
      To work around this, use HTTP or write your own certificate verification
      logic. We recommend that you do not use periods (".") in bucket names when
      using virtual hosted-style buckets.

    The following are the rules for naming S3 keys:

    - The name for a key is a sequence of Unicode characters whose UTF-8
      encoding is at most 1024 bytes long.

    :param bucket: str for bucket in :code:`s3://{bucket}/{key}`
    :param key: str for key in :code:`s3://{bucket}/{key}`
    """

    @staticmethod
    def parse_s3_uri(s3_uri: str) -> "S3URI":
        s3_paths = S3Paths.parse_s3_uri(s3_uri)
        return S3URI(bucket=s3_paths.bucket, key=s3_paths.key)

    def s3_bucket(self) -> "s3.Bucket":
        """A :code:`boto3.resource('s3').Bucket('bucket')`"""
        return boto3.resource("s3").Bucket(self.bucket)

    def s3_derivatives(self) -> Iterable:
        """Use :code:`s3_objects(glob_file_pattern)` to find all the derivatives
        with matching file names (key_file*.*) below the path of this file.

        :return: Iterable[s3.ObjectSummary]
        """
        pattern = self.glob_file_pattern()
        return self.s3_objects(glob_pattern=pattern)

    def s3_exists(self) -> bool:
        """Check the s3 file summary to confirm a file exists.
        :return: bool True if the s3 file exists
        """
        if self.s3_last_modified():
            return True
        return False

    def s3_head_request(self) -> dict:
        """Issue an HTTP HEAD request to get the s3 URI summary data.

        :return: dict of HEAD response data or an empty dict if it fails.
        """
        try:
            # check that an s3 key exists with a head request
            client = boto3.client("s3")
            return client.head_object(Bucket=self.bucket, Key=self.key)

        except botocore.exceptions.ClientError as err:
            self._log_client_errors(err)

        return {}

    def s3_last_modified(self) -> Union[datetime, None]:
        try:
            return self.s3_object_summary().last_modified

        except botocore.exceptions.ClientError as err:
            self._log_client_errors(err)

        return None

    def s3_object_summary(self) -> "s3.ObjectSummary":
        """Create an s3.ObjectSummary for this S3URI;
        this does not check if the object exists.

        :return: s3.ObjectSummary
        """
        s3 = boto3.resource("s3")
        return s3.ObjectSummary(self.bucket, self.key)

    def s3_objects(self, glob_pattern: str = None, prefix: str = None) -> Iterable:
        """
        Search all the s3 objects at or below the s3_uri.bucket, using an
        optional filter prefix and/or a glob pattern (defaults to everything).

        Using the s3_uri.key_path as a filter prefix is most often critical
        to limit the number of bucket objects to process. The addition of a
        glob pattern is applied to the results of all objects found; the glob
        pattern match is applied to each s3.ObjectSummary.key.

        :param prefix: str to specify a Prefix filter based on glob patterns
        :param glob_pattern: str to specify a filter based on glob patterns
        :return: Iterable[s3.ObjectSummary]
        """
        s3_bucket = self.s3_bucket()
        if prefix:
            objects = s3_bucket.objects.filter(Prefix=prefix)
        else:
            objects = s3_bucket.objects.all()
        if glob_pattern:
            objects = (
                s3_obj
                for s3_obj in objects
                if PurePosixPath(s3_obj.key).match(glob_pattern)
            )
        return objects

    def _log_client_errors(self, err):
        # Note: this does _not_ use an instance variable for self.logger
        #       because the loggers cannot be pickled
        status = err.response["ResponseMetadata"]["HTTPStatusCode"]
        errcode = err.response["Error"]["Code"]
        if status == 404:
            LOGGER.warning("Missing object, %s for %s", errcode, self.s3_uri)
        elif status == 403:
            LOGGER.error("Access denied, %s for %s", errcode, self.s3_uri)
        else:
            LOGGER.exception("Error in request, %s for %s", errcode, self.s3_uri)


@dataclass
class S3Info:
    s3_uri: S3URI
    s3_size: int = None
    last_modified: datetime = None

    @property
    def iso8601(self):
        if self.last_modified:
            return self.last_modified.isoformat()

    @property
    def dict(self) -> Dict:
        return {
            "s3_uri": str(self.s3_uri),
            "s3_size": self.s3_size,
            "last_modified": self.iso8601,
        }

    @property
    def json(self) -> str:
        return json.dumps(self.dict)



# Copyright 2020 Darren Weber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
AioAWS logger
"""

import logging
import os
import sys
import time

logging.Formatter.converter = time.gmtime

LOG_LEVEL = os.environ.get("LOG_LEVEL", "INFO").upper()
LOG_FORMAT = "[%(levelname)s]  %(asctime)s.%(msecs)03dZ  %(name)s:%(funcName)s:%(lineno)d  %(message)s"
LOG_FORMATTER = logging.Formatter(LOG_FORMAT, "%Y-%m-%dT%H:%M:%S")
HANDLER = logging.StreamHandler(sys.stdout)
HANDLER.formatter = LOG_FORMATTER


def get_logger(name: str = "aio-aws") -> logging.Logger:
    logger = logging.getLogger(name)
    if not any(isinstance(h, logging.StreamHandler) for h in logger.handlers):
        logger.addHandler(HANDLER)
    logger.setLevel(LOG_LEVEL)
    logger.propagate = False
    return logger



#! /usr/bin/env python3
# pylint: disable=bad-continuation

# Copyright 2020 Darren Weber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Asyncio Executor
----------------

This module explores an AsyncioExecutor, which follows the
execution API of ``concurrent.futures.Executor``.

Example main script to test asyncio task completions
using an :py:class:`.AsyncioExecutor`

.. code-block:: shell

    $ ./aio_aws/async_executor.py --task-count 5
    5 async tasks submitted in 0.00 seconds.
    Task 0 - await a sleep for 6.86
    Task 1 - await a sleep for 4.50
    Task 2 - await a sleep for 3.31
    Task 3 - await a sleep for 6.41
    Task 4 - await a sleep for 3.06
    Task 4 - done with sleep for 3.06
    Task 2 - done with sleep for 3.31
    Task 1 - done with sleep for 4.50
    Task 3 - done with sleep for 6.41
    Task 0 - done with sleep for 6.86
    5 async tasks completed in 6.87 seconds.
    5 async tasks validated in 6.87 seconds.

    $ ./aio_aws/async_executor.py --task-count 10000
    10000 async tasks submitted in 0.01 seconds.
    Task 0 - await a sleep for 8.39
    Task 1 - await a sleep for 6.96
    Task 2 - await a sleep for 2.07
    ...
    Task 916 - done with sleep for 1.05
    Task 15 - done with sleep for 1.19
    Task 1118 - done with sleep for 1.04
    ...
    Task 9831 - done with sleep for 9.96
    Task 9687 - done with sleep for 9.97
    Task 9872 - done with sleep for 9.99
    10000 async tasks completed in 12.44 seconds.
    10000 async tasks validated in 12.47 seconds.


.. note::
    Code in this module is partly copied and adapted from:

    - https://gist.github.com/seglberg/0b4487b57b4fd425c56ad72aba9971be
    - public-domain license (as of Jan 2020)

"""

import asyncio
import concurrent.futures
import inspect
import random
import threading
import time
from asyncio import Future
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Coroutine
from typing import Iterable
from typing import Iterator
from typing import Optional
from typing import Union

from aio_aws.logger import get_logger

LOGGER = get_logger(__name__)


def loop_mgr(loop: asyncio.AbstractEventLoop):
    """
    An asyncio loop manager, used by :py:class:`.AsyncioExecutor`
    to run the loop forever in a thread and clean up after the
    loop stops.

    :param loop:
    """
    try:
        # loop manager will run this in it's own thread
        loop.run_forever()

        # the loop was stopped and concurrent.futures.Executor
        # promises to complete tasks on shutdown.
        while True:
            tasks = asyncio.all_tasks(loop=loop)
            pending = [t for t in tasks if not t.done()]
            loop.run_until_complete(asyncio.gather(*pending))

            # ensure the task collection is updated
            # (this is _not_ redundant)
            tasks = asyncio.all_tasks(loop=loop)
            if all([t.done() for t in tasks]):
                break

    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        # AsyncioExecutor might restart the loop, so skip:
        # loop.close()  # irreversible


class AsyncioExecutor(concurrent.futures.Executor):
    """
    AsyncioExecutor follows the ``concurrent.futures.Executor`` API

    It wraps an ``asyncio.AbstractEventLoop`` in a thread to manage
    submission of coroutines and blocking tasks.

    Examples:

    .. code-block::

        async def delay(pause: int) -> float:
            try:
                await asyncio.sleep(pause)
                return pause
            except asyncio.CancelledError:
                raise

        loop_executor = AsyncioExecutor()
        task = loop_executor.submit(delay, 1)
        loop_executor.shutdown()  # waits for tasks to complete
        assert task.done()
        pause = task.result()
        assert pause == 1

    .. seealso::
        - https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.Executor

    """

    def __init__(self):

        super().__init__()
        # getting an event loop should only be done once, here, but the loop property
        # exposes the loop and must check if something has closed the loop to get a new one.
        self._loop = asyncio.new_event_loop()
        self._shutdown = False
        self._thread = None
        self.start()

        # Consider managing blocking functions with a process executor, although this
        # might really confuse the purpose and responsibility of this class; so PUNT.
        # self._process_executor = concurrent.futures.ProcessPoolExecutor()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop and self._loop.is_closed():
            # Note: not exactly sure how this interacts with start() and shutdown().
            self._loop = asyncio.new_event_loop()
        return self._loop

    def is_alive(self) -> bool:
        if self._shutdown:
            return False
        return self._thread and self._thread.is_alive() and self.loop.is_running()

    def start(self) -> bool:
        # Use a loop manager to run tasks in it's own thread

        # executors own a new event loop and run it
        # in their own thread; not sure what the limits are to
        # how many executor-loops can be spawned, but the class
        # does not use any kind of thread pool (yet).

        if self.is_alive():
            return True

        if self._shutdown and self._thread and self._thread.is_alive():
            while self._thread.is_alive():
                time.sleep(0.2)
        self._shutdown = False
        self._thread = threading.Thread(target=loop_mgr, args=(self.loop,), daemon=True)
        self._thread.start()
        while not self._thread.is_alive():
            time.sleep(0.2)
        return self.is_alive()

    def map(
        self,
        func: Callable,
        *iterables: Iterable[Any],
        timeout: Optional[float] = 120,
        chunksize: int = 100,
    ) -> Iterator[Future]:
        """TODO docs"""

        self._check_loop()

        # TODO: apply chunksize to iteration?  This might be done with a
        #       asyncio.Semaphore but that requires an async coroutine.
        async_tasks = []
        for obj in iterables:
            coro = func(obj)
            # coroutine objects can be wrapped in an async task (future);
            # the event loop creates the task, but does not start it until
            # this creation coroutine is awaited (run by event loop).
            async_task = self.loop.create_task(coro)
            async_tasks.append(async_task)

        return asyncio.as_completed(async_tasks, loop=self.loop, timeout=timeout)

    def submit(
        self, fn: Union[Callable, Coroutine, Awaitable], *args, **kwargs
    ) -> Optional[Future]:
        """
        Schedules the callable, fn, to be executed as ``fn(*args **kwargs)`` and
        returns a Future object representing the execution of the callable.

        Add aio_aws for coroutine-functions, coroutine-objects, and blocking functions;
        apply the right typing annotations.  Maybe split this function up into different
        public (or private) methods to handle different fn types?

        :param fn:
        :param args:
        :param kwargs:
        :return:
        """

        self._check_loop()

        if inspect.iscoroutinefunction(fn):
            coro = fn(*args, **kwargs)
            return asyncio.run_coroutine_threadsafe(coro, self.loop)

        if inspect.iscoroutine(fn):
            return asyncio.run_coroutine_threadsafe(fn, self.loop)

        if inspect.isawaitable(fn) and inspect.isgenerator(fn):
            return asyncio.run_coroutine_threadsafe(fn, self.loop)

        if inspect.isgeneratorfunction(fn):
            gen = fn(*args, **kwargs)
            if inspect.isawaitable(gen) and inspect.isgenerator(gen):
                return asyncio.run_coroutine_threadsafe(gen, self.loop)

        raise RuntimeError("Can only execute awaitable coroutines and generators")

        # Try to submit blocking functions to a process executor, although this
        # might really confuse the purpose and responsibility of this class.
        # func = functools.partial(fn, *args, **kwargs)
        # return self.loop.run_in_executor(self._process_executor, func)

    def shutdown(self, wait=True):
        """
        https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.Executor.shutdown

        concurrent.futures.Executor promises to complete
        tasks on shutdown, so gather all pending tasks.

        promises that submitted tasks will run to completion (not cancelled); maybe add
        aio_aws about how the user can first cancel tasks and then shutdown this executor
        :param wait: wait for thread running event loop to join
        """
        if self._shutdown:
            LOGGER.warning("Already shutdown")
            return

        # for some reason, if nothing has been submitted, the shutdown hangs on a thread lock,
        # so first submit a small async sleep to work around this until it can be fixed.
        asyncio.run_coroutine_threadsafe(asyncio.sleep(0.01), self.loop)

        self._shutdown = True
        self.loop.stop()
        if wait:
            while self.loop.is_running():
                pass
            self._thread.join()
        self._thread = None

    def _check_loop(self):
        if self._shutdown:
            raise RuntimeError("Cannot schedule new futures after shutdown")

        if not self.is_alive():
            raise RuntimeError(
                "Loop must be started before any function can be submitted"
            )


#
#  main test code
#

import click

#: Minimum task pause
MIN_PAUSE: int = 1

#: Maximum task pause
MAX_PAUSE: int = 10


async def delay(task_id: int) -> float:
    """
    Await a random pause between :py:const:`MIN_PAUSE` and :py:const:`MAX_PAUSE`

    :param task_id: the ID for the asyncio.Task awaiting this pause
    :return: random interval for pause
    """
    pause = random.uniform(MIN_PAUSE, MAX_PAUSE)
    LOGGER.warning("Task %d - await a sleep for %.2f", task_id, pause)
    try:
        await asyncio.sleep(pause)
        LOGGER.warning("Task %d - done with sleep for %.2f", task_id, pause)
        return pause

    except asyncio.CancelledError:
        LOGGER.error("Task %d - cancelled", task_id)
        raise


async def submit_throttle(
    executor: AsyncioExecutor,
    sem: asyncio.Semaphore,
    fn: Union[Callable, Coroutine, Awaitable],
    *args,
    **kwargs,
) -> Optional[Future]:

    # AsyncioExecutor has no async API so it can't ow this function as a method,
    # which is a real problem for using the concurrent.futures.Executor API.  So
    # this function needs to be called in a loop that is not owned by the executor.
    # That might raise issues with thread-safe calls.

    async with sem:
        task = executor.submit(fn, *args, **kwargs)
        return task


@click.command()
@click.option(
    "--task-count",
    default=5,
    show_default=True,
    type=int,
    help="A number of tasks to run",
)
def main(task_count):
    """Run asyncio executor"""

    loop_executor = AsyncioExecutor()
    # task_limits = asyncio.Semaphore(50)  # limit to 50 tasks

    tasks = []
    start = time.perf_counter()
    try:
        # Run a small set of simple asyncio tasks that pause up to 10 sec
        for task_id in range(task_count):
            task = loop_executor.submit(delay, task_id)
            # task = submit_throttle(loop_executor, task_limits, delay, task_id)
            tasks.append(task)
        end = time.perf_counter() - start
        print(f"{len(tasks):d} async tasks submitted in {end:0.2f} seconds.")

        # # It's possible to add callbacks while the event loop continues to run;
        # # bit it complicates shutdown, where callbacks are not guaranteed.
        # for task in tasks:
        #     task.add_done_callback(partial(print, "Future:"))

        # Collect tasks while loop is running; skip it by default to test shutdown.
        collect = False
        if collect:
            loop_executor.loop.run_until_complete(asyncio.gather(*tasks))
            for task in tasks:
                assert task.done()  # check the task completed
                assert task.exception() is None
                task_pause = task.result()
                assert isinstance(task_pause, float)

        loop_executor.shutdown()  # this must wait for tasks to complete
        assert not loop_executor.loop.is_running()
        end = time.perf_counter() - start
        print(f"{len(tasks):d} async tasks completed in {end:0.2f} seconds.")

        # Test that shutdown allowed all tasks to complete.
        for task in tasks:
            assert task.done()  # check the executor ran to completion
            assert task.exception() is None
            task_pause = task.result()
            assert isinstance(task_pause, float)
        end = time.perf_counter() - start
        print(f"{len(tasks):d} async tasks validated in {end:0.2f} seconds.")

    finally:
        loop = loop_executor.loop
        while loop.is_running():
            loop.shutdown_asyncgens()
            loop.stop()
        loop.close()
        assert not loop.is_running()
        assert loop.is_closed()


if __name__ == "__main__":

    main()  # pylint: disable=no-value-for-parameter



#! /usr/bin/env python3

# Copyright 2020 Darren Weber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
AioAWS Lambda
-------------

Manually create a 'lambda_dev' function with simple code like:

.. code-block::

    # lambda_dev/lambda_function.py

    import json

    def lambda_handler(event, context):
        csv = ", ".join([str(i) for i in range(10)])
        print(csv)  # to log something
        return {
            'statusCode': 200,
            'body': json.dumps(csv)
        }

The `aio_aws_lambda.py` module has a main script to call this simple lambda function.
This lambda function easily runs within a 128 Mb memory allocation and it runs in
less time than the minimal billing period, so it's as cheap as it can be and could
fit within a monthly free quota.

.. seealso::
    - https://aws.amazon.com/lambda/pricing/

.. code-block::

    $ ./aio_aws/aio_aws_lambda.py
    Test async lambda functions
    [INFO]  2020-04-07T03:10:21.741Z  aio-aws:aio_lambda_invoke:114  AWS Lambda (lambda_dev) invoked OK
    1 lambdas finished in 0.31 seconds.

    $ N_LAMBDAS=1000 ./aio_aws/aio_aws_lambda.py
    # snipped logging messages
    1000 lambdas finished in 14.95 seconds.

"""

import asyncio
import base64
import concurrent.futures
import json
import os
import time
from dataclasses import dataclass
from functools import partial
from json import JSONDecodeError
from typing import Dict
from typing import List
from typing import Optional

import aiobotocore.client
import botocore.exceptions

from aio_aws.aio_aws_config import AioAWSConfig
from aio_aws.aio_aws_config import jitter
from aio_aws.aio_aws_config import MAX_POOL_CONNECTIONS
from aio_aws.utils import response_success
from aio_aws.logger import get_logger

LOGGER = get_logger(__name__)

MAXIMUM_PAYLOAD_SIZE = 6291556


@dataclass
class AWSLambdaFunction:
    """
    AWS Lambda Function

    Creating an AWSLambdaFunction instance does not create or invoke anything, it's
    a dataclass to retain and track function parameters and response data.

    :param name: A lambda FunctionName (truncated to 64 characters).
    :param type: the type of function invocation ("RequestResponse", "Event", "DryRun")
    :param log_type: the type of function logging ("None" or "Tail")
    :param context: the client context; for an example of a ClientContext JSON,
        see PutEvents in the Amazon Mobile Analytics API Reference and User Guide. The
        ClientContext JSON must be base64-encoded and has a maximum size of 3583 bytes.
    :param payload: an input payload (bytes or seekable file-like object); JSON that you
        want to provide to your Lambda function as input.
    :param qualifier: an optional function version or alias name;
        If you specify a function version, the API uses the qualified
        function ARN to invoke a specific Lambda function. If you specify an alias
        name, the API uses the alias ARN to invoke the Lambda function version to
        which the alias points. If you don't provide this parameter, then the API
        uses unqualified function ARN which results in invocation of the $LATEST version.
    """

    name: str
    type: str = "RequestResponse"  # or "Event" or "DryRun"
    log_type: str = "None"  # "None" or "Tail"
    context: str = None
    payload: bytes = b""  # or a file
    qualifier: str = None
    response: Dict = None
    data: bytes = None

    TYPES = ["RequestResponse", "Event", "DryRun"]
    LOG_TYPES = ["None", "Tail"]

    def __post_init__(self):

        self.name = self.name[:64]

        if self.type not in self.TYPES:
            raise ValueError(f"The type ({self.type}) must be one of {self.TYPES}")

        if self.type != "RequestResponse":
            self.log_type = "None"
        elif self.log_type not in self.LOG_TYPES:
            raise ValueError(
                f"The log_type ({self.log_type}) must be one of {self.LOG_TYPES}"
            )

    @property
    def response_metadata(self) -> Optional[Dict]:
        if self.response:
            return self.response.get("ResponseMetadata")

    @property
    def response_headers(self) -> Optional[Dict]:
        metadata = self.response_metadata
        if metadata:
            return metadata.get("HTTPHeaders")

    @property
    def content_type(self) -> Optional[str]:
        headers = self.response_headers
        if headers:
            return headers.get("content-type")

    @property
    def content_length(self) -> Optional[int]:
        headers = self.response_headers
        if headers:
            return int(headers.get("content-length"))

    @property
    def status_code(self) -> Optional[int]:
        if self.response:
            return self.response.get("StatusCode")

    @property
    def error(self):
        """This could be a lambda function error or a client error"""
        if self.response:
            if self.response.get("FunctionError"):
                # lambda function error
                if self.data:
                    error = self.data.decode()
                    try:
                        error = json.loads(error)
                    except JSONDecodeError:
                        pass
                    return error

            if self.response.get("Error"):
                # boto3 client error
                return self.response.get("Error")

    @property
    def text(self):
        if self.response and not self.error:
            if self.data:
                return self.data.decode()

    @property
    def json(self):
        text = self.text
        if text:
            return json.loads(text)

    @property
    def logs(self):
        if self.response:
            log_result = self.response.get("LogResult")
            if log_result:
                return base64.b64decode(log_result)

    @property
    def params(self):
        """AWS Lambda parameters to invoke function"""
        params = {
            "FunctionName": self.name,
            "InvocationType": self.type,
            "LogType": self.log_type,
        }
        if self.context:
            params["ClientContext"] = self.context
        if self.payload:
            params["Payload"] = self.payload
        if self.qualifier:
            params["Qualifier"] = self.qualifier
        return params

    async def invoke(
        self, config: AioAWSConfig, lambda_client: aiobotocore.client.AioBaseClient
    ) -> "AWSLambdaFunction":
        """
        Asynchronous coroutine to invoke a lambda function; this
        updates the ``response`` and calls the py:meth:`.read_response`
        method to handle the response.

        :param config: aio session and client settings
        :param lambda_client: aio client for lambda
        :return: a lambda response
        :raises: botocore.exceptions.ClientError, botocore.exceptions.ParamValidationError
        """
        async with config.semaphore:
            for tries in range(config.retries + 1):
                try:
                    LOGGER.debug("AWS Lambda params: %s", self.params)
                    response = await lambda_client.invoke(**self.params)
                    self.response = response
                    LOGGER.debug("AWS Lambda response: %s", self.response)

                    if response_success(self.response):
                        await self.read_response()  # updates self.data
                        if self.data:
                            LOGGER.info("AWS Lambda invoked OK: %s", self.name)
                        else:
                            error = self.error
                            if error:
                                LOGGER.error(
                                    "AWS Lambda error: %s, %s", self.name, error
                                )
                    else:
                        # TODO: are there some failures that could be recovered here?
                        LOGGER.error("AWS Lambda invoke failure: %s", self.name)

                    return self

                except botocore.exceptions.ClientError as err:
                    response = err.response
                    LOGGER.error("AWS Lambda client error: %s, %s", self.name, response)
                    error = response.get("Error", {})
                    if error.get("Code") == "TooManyRequestsException":
                        if tries < config.retries:
                            await jitter(
                                "lambda-retry", config.min_jitter, config.max_jitter
                            )
                        continue  # allow it to retry, if possible
                    else:
                        self.response = response
                        raise

            raise RuntimeError("AWS Lambda invoke exceeded retries")

    async def read_response(self):
        """
        Asynchronous coroutine to read a lambda response; this
        updates the ``data`` attribute.

        :raises: botocore.exceptions.ClientError, botocore.exceptions.ParamValidationError
        """
        if self.response and response_success(self.response):
            response_payload = self.response.get("Payload")
            if response_payload:
                async with response_payload as stream:
                    self.data = await stream.read()


async def run_lambda_functions(lambda_functions: List[AWSLambdaFunction]):
    """Use some default config settings to run lambda functions

    .. code-block::

        lambda_funcs = []
        for i in range(5):
            event = {"i": i}
            payload = json.dumps(event).encode()
            func = AWSLambdaFunction(name="lambda_dev", payload=payload)
            lambda_funcs.append(func)
        asyncio.run(run_lambda_functions(lambda_funcs))
        for func in lambda_funcs:
            assert response_success(func.response)

    :return: it returns nothing, the lambda function will contain the
        results of any lambda response in func.response and func.data
    """
    config = AioAWSConfig(
        retries=0,
        aws_region=os.getenv("AWS_DEFAULT_REGION", "us-west-2"),
        min_jitter=0.2,
        max_jitter=0.8,
        max_pool_connections=MAX_POOL_CONNECTIONS,
    )
    lambda_tasks = []
    async with config.create_client("lambda") as lambda_client:
        for lambda_func in lambda_functions:
            lambda_task = asyncio.create_task(lambda_func.invoke(config, lambda_client))
            lambda_tasks.append(lambda_task)
        await asyncio.gather(*lambda_tasks)


async def run_lambda_function_thread_pool(
    lambda_functions: List[AWSLambdaFunction], n_tasks: int = 4
):
    config = AioAWSConfig(
        retries=0,
        aws_region=os.getenv("AWS_DEFAULT_REGION", "us-west-2"),
        min_jitter=0.2,
        max_jitter=0.8,
        max_pool_connections=MAX_POOL_CONNECTIONS,
    )

    with concurrent.futures.ThreadPoolExecutor() as executor:
        loop = asyncio.get_running_loop()
        lambda_tasks = []

        async with config.create_client("lambda") as lambda_client:
            for lambda_func in lambda_functions:
                func = partial(
                    lambda_func.invoke,
                    config=config,
                    lambda_client=lambda_client,
                )
                lambda_task = await loop.run_in_executor(executor=executor, func=func)
                lambda_tasks.append(lambda_task)

                # Limit concurrency to some extent
                if len(lambda_tasks) == n_tasks:
                    await asyncio.gather(*lambda_tasks)
                    lambda_tasks = []

            # collect any remaining tasks in the context of the thread pool
            await asyncio.gather(*lambda_tasks)


if __name__ == "__main__":

    # This __main__ code can be useful to test against AWS Lambda
    # because the moto pytest suite is not an exact replica.  This
    # requires a few minutes to setup a live AWS Lambda function
    # called 'lambda_dev' - see the test suite for a sample handler.

    from pprint import pprint

    print()
    print("Test async lambda functions")
    start = time.perf_counter()
    N_lambdas = int(os.environ.get("N_LAMBDAS", 1))

    lambda_funcs = []
    for i in range(N_lambdas):
        event = {"i": i}
        # event = {"action": "too-large"}
        # event = {"action": "runtime-error"}
        payload = json.dumps(event).encode()
        func = AWSLambdaFunction(name="lambda_dev", payload=payload)
        lambda_funcs.append(func)

    asyncio.run(run_lambda_functions(lambda_funcs))
    # # Note: a thread pool executor is not faster than asyncio alone
    # asyncio.run(run_lambda_function_thread_pool(lambda_funcs, n_tasks=N_lambdas))

    responses = []
    for func in lambda_funcs:
        assert response_success(func.response)
        if N_lambdas < 3:
            print()
            print("Params:")
            pprint(func.params)
            print("Response:")
            pprint(func.response)
            print("Data:")
            pprint(func.data)
            print("JSON:")
            pprint(func.json)
            print("Error:")
            pprint(func.error)
            print("Logs")
            pprint(func.logs)
        elif N_lambdas < 20:
            responses.append(func.json)

    if responses:
        # print(json.dumps(responses, indent=2))
        pprint(responses)

    end = time.perf_counter() - start
    print()
    print(f"{N_lambdas} lambdas finished in {end:0.2f} seconds.\n")



"""
S3 AIO
======

These functions provide non-blocking serialization to files
and s3 for GeoJSON, GeoJSONSeq, JSON, and YAML.

"""

import asyncio
import json
import os
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import aiofiles
import yaml
from aiobotocore.client import AioBaseClient
from aiofiles.tempfile import NamedTemporaryFile
from botocore.exceptions import ClientError

from aio_aws.aio_aws_config import aio_aws_client
from aio_aws.logger import get_logger
from aio_aws.s3_uri import S3Info
from aio_aws.s3_uri import S3URI
from aio_aws.utils import response_success

LOGGER = get_logger(__name__)


@asynccontextmanager
async def s3_aio_client(*args, **kwargs) -> AioBaseClient:
    """
    This creates a aiobotocore.client.AioBaseClient that uses aiohttp
    for asynchronous s3 requests

    :return: a aiobotocore.client.AioBaseClient for s3
    """
    async with aio_aws_client("s3", *args, **kwargs) as s3_client:
        yield s3_client


async def s3_file_info(s3_uri: Union[S3URI, str], s3_client: AioBaseClient) -> S3Info:
    """
    Collect data from an S3 HEAD request for an S3URI

    :param s3_uri: a fully qualified S3 URI for the s3 object to read
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: an S3Info object with HEAD data on success;
        on failure the S3Info object has no HEAD data
    """
    if isinstance(s3_uri, str):
        s3_uri = S3URI(s3_uri)

    s3_info = S3Info(s3_uri=s3_uri)
    try:
        s3_head = await s3_client.head_object(Bucket=s3_uri.bucket, Key=s3_uri.key)
        if response_success(s3_head):
            # LastModified is a datetime.datetime
            s3_info.last_modified = s3_head["LastModified"]
            s3_info.s3_size = int(s3_head["ContentLength"])
            LOGGER.debug("Success S3URI info: %s", s3_uri)
    except ClientError as err:
        LOGGER.debug("Failed S3URI info: %s", s3_uri)
        LOGGER.debug(err)
    return s3_info


async def s3_files_info(
    s3_uris: List[Union[S3URI, str]], s3_client: AioBaseClient
) -> List[S3Info]:
    """
    Collect data from S3 HEAD requests for many S3URI

    :param s3_uris: a list of S3URI
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: a list of S3Info object with HEAD data on success;
        on failure the S3Info object has no HEAD data
    """
    s3_files = []
    s3_futures = []
    for s3_uri in s3_uris:
        s3_future = asyncio.ensure_future(s3_file_info(s3_uri, s3_client))
        s3_futures.append(s3_future)
    for future in asyncio.as_completed(s3_futures):
        try:
            s3_info = await future
            s3_files.append(s3_info)
        except Exception as err:
            LOGGER.error(err)
            pass

    return s3_files


async def get_s3_content(s3_uri: str, s3_client: AioBaseClient):
    """
    Read an s3 object

    :param s3_uri: a fully qualified S3 URI for an s3 object
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: the data from the s3 object
    """
    try:
        s3_uri = S3URI(s3_uri)
        LOGGER.info("Read s3-uri: %s", s3_uri.s3_uri)
        content_object = await s3_client.get_object(
            Bucket=s3_uri.bucket, Key=s3_uri.key
        )
        file_content = await content_object["Body"].read()
        return file_content.decode("utf-8")
    except ClientError as err:
        LOGGER.error("Failed S3 GET for: %s", s3_uri)
        LOGGER.error(err)


async def put_s3_content(
    data_file: str, s3_uri: str, s3_client: AioBaseClient
) -> Optional[str]:
    """
    Write a file to an s3 object

    :param data_file: a data file
    :param s3_uri: a fully qualified S3 URI for an s3 object
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: the s3 URI on success
    """
    s3_uri = S3URI(s3_uri)
    try:
        async with aiofiles.open(data_file, "rb") as fd:
            file_bytes = await fd.read()
            response = await s3_client.put_object(
                Bucket=s3_uri.bucket, Key=s3_uri.key, Body=file_bytes
            )
            success = response_success(response)
            if success:
                exists_waiter = s3_client.get_waiter("object_exists")
                await exists_waiter.wait(Bucket=s3_uri.bucket, Key=s3_uri.key)
                return str(s3_uri)
    except ClientError as err:
        LOGGER.error("Failed S3 PUT to: %s", s3_uri)
        LOGGER.error(err)


async def json_dump(data: Any, file: Union[Path, str]) -> Optional[Union[Path, str]]:
    """
    Write JSON to a file

    :param data: any data compatible with json.dumps
    :param file: a file path to write
    :return: if the dump succeeds, return the file, or None
    """
    dump = json.dumps(data)
    async with aiofiles.open(file, mode="w") as dst:
        await dst.write(dump)
    file_path = Path(file)
    if file_path.is_file() and file_path.stat().st_size > 0:
        LOGGER.info("Saved JSON to %s", file)
        return file
    LOGGER.error("Failed to save JSON to %s", file)


async def json_s3_dump(
    json_data: Any, s3_uri: str, s3_client: AioBaseClient
) -> Optional[str]:
    """
    Write JSON to an s3 URI

    :param json_data: an object to json.dump
    :param s3_uri: a fully qualified S3 URI for the s3 object to write
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: the s3 URI on success
    """
    s3_uri = S3URI(s3_uri)
    success = False
    tmp_file = None
    try:
        async with NamedTemporaryFile(delete=False) as o_file:
            tmp_file = o_file.name
            async with aiofiles.open(o_file.name, "w") as fd:
                dumps = json.dumps(json_data)
                await fd.write(dumps)

        s3_obj = await put_s3_content(
            data_file=tmp_file, s3_uri=str(s3_uri), s3_client=s3_client
        )
        if s3_obj:
            success = True

    finally:
        if tmp_file:
            os.unlink(tmp_file)

    if success:
        LOGGER.info("Saved S3URI: %s", str(s3_uri))
        return str(s3_uri)
    else:
        LOGGER.error("Failed to save S3URI: %s", str(s3_uri))


async def json_s3_load(s3_uri: str, s3_client: AioBaseClient) -> Any:
    """
    Read JSON data from an s3 object

    :param s3_uri: a fully qualified S3 URI for an s3 object
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: data from the json load
    """
    file_content = await get_s3_content(s3_uri, s3_client=s3_client)
    json_data = json.loads(file_content)
    return json_data


async def geojson_s3_load(s3_uri: str, s3_client: AioBaseClient) -> Dict:
    """
    Read GeoJSON data from an s3 object

    :param s3_uri: a fully qualified S3 URI for an s3 object
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: geojson data
    """
    file_content = await get_s3_content(s3_uri, s3_client=s3_client)
    geojson = json.loads(file_content)
    return geojson


async def geojson_s3_dump(
    geojson_data: Any, s3_uri: str, s3_client: AioBaseClient
) -> Optional[str]:
    """
    Write GeoJSON to an s3 URI

    :param geojson_data: an object to json.dump
    :param s3_uri: a fully qualified S3 URI for an s3 object
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: the s3 URI on success
    """
    return await json_s3_dump(
        json_data=geojson_data, s3_uri=s3_uri, s3_client=s3_client
    )


async def geojsons_s3_load(s3_uri: str, s3_client: AioBaseClient) -> List[Dict]:
    """
    Read GeoJSON Text Sequence data from an s3 object

    :param s3_uri: a fully qualified S3 URI for an s3 object
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: geojson features

    .. seealso::

        - https://tools.ietf.org/html/rfc8142

    """
    file_content = await get_s3_content(s3_uri, s3_client=s3_client)
    geojsons = file_content.splitlines()
    # some of the geojsons lines could be empty
    features = []
    while geojsons:
        feature = geojsons.pop(0).strip()
        if feature:
            features.append(json.loads(feature))
    return features


async def geojsons_s3_dump(
    geojson_features: List[Dict], s3uri: str, s3_client: AioBaseClient
) -> Optional[str]:
    """
    Write GeoJSON Text Sequence files to an s3 URI

    [GeoJSON Text Sequences](https://tools.ietf.org/html/rfc8142) are
    lines of geojson features that are designed for streaming
    operations on large datasets. These files can be loaded by
    geopandas, using fiona `driver="GeoJSONSeq"`, which
    can be auto-detected.  For example:

    .. code-block::

        import geopandas as gpd

        s3_uri = "s3://your-bucket/prefix/input.geojsons"
        gdf = gpd.read_file(s3_uri)

    :param geojson_features: a list of geojson features; from any
        feature collection, this is geojson_collection["features"]
    :param s3uri: a fully qualified S3 URI for an s3 object
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: the s3 URI on success
    """
    s3_uri = S3URI(s3uri)
    success = False
    tmp_file = None
    try:
        async with NamedTemporaryFile(delete=False) as o_file:
            tmp_file = o_file.name
            async with aiofiles.open(o_file.name, "w") as fd:
                for feature in geojson_features:
                    dumps = json.dumps(feature)
                    await fd.write(dumps + "\n")

        s3_obj = await put_s3_content(
            data_file=tmp_file, s3_uri=str(s3_uri), s3_client=s3_client
        )
        if s3_obj:
            success = True

    finally:
        if tmp_file:
            os.unlink(tmp_file)

    if success:
        LOGGER.info("Saved GeoJSONSeq to %s", str(s3_uri))
        return str(s3_uri)
    else:
        LOGGER.error("Failed to save GeoJSONSeq to %s", s3_uri)


async def geojsons_dump(
    geojson_features: List[Dict], geojsons_file: Union[Path, str]
) -> Optional[Union[Path, str]]:
    """
    Write a GeoJSON Text Sequence file

    :param geojson_features: a list of geojson features; from any
        feature collection, this is geojson_collection["features"]
    :param geojsons_file: a file path to write
    :return: if the dump succeeds, return the geojsons_file, or None
    """
    async with aiofiles.open(geojsons_file, mode="w") as dst:
        for feature in geojson_features:
            dump = json.dumps(feature)
            await dst.write(dump + "\n")
    geojsons_path = Path(geojsons_file)
    if geojsons_path.is_file() and geojsons_path.stat().st_size > 0:
        LOGGER.info("Saved GeoJSONSeq to %s", geojsons_file)
        return geojsons_file
    LOGGER.error("Failed to save GeoJSONSeq to %s", geojsons_file)


async def yaml_s3_dump(
    yaml_data: Any, s3_uri: str, s3_client: AioBaseClient
) -> Optional[str]:
    """
    Write YAML to an s3 URI

    :param yaml_data: an object to yaml.dump
    :param s3_uri: a fully qualified S3 URI for an s3 object
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: the s3 URI on success
    """
    s3_uri = S3URI(s3_uri)
    success = False
    tmp_file = None
    try:
        async with NamedTemporaryFile(delete=False) as o_file:
            tmp_file = o_file.name
            async with aiofiles.open(o_file.name, "w") as fd:
                dumps = yaml.dump(yaml_data)
                await fd.write(dumps)

        s3_obj = await put_s3_content(
            data_file=tmp_file, s3_uri=str(s3_uri), s3_client=s3_client
        )
        if s3_obj:
            success = True

    finally:
        if tmp_file:
            os.unlink(tmp_file)

    if success:
        LOGGER.info("Saved S3URI: %s", str(s3_uri))
        return str(s3_uri)
    else:
        LOGGER.error("Failed to save S3URI: %s", str(s3_uri))


async def yaml_s3_load(s3_uri: str, s3_client: AioBaseClient) -> Any:
    """
    Read YAML data from an s3 object

    :param s3_uri: a fully qualified S3 URI for an s3 object
    :param s3_client: a required aiobotocore.client.AioBaseClient for s3
    :return: data from the yaml load
    """
    file_content = await get_s3_content(s3_uri, s3_client=s3_client)
    yaml_data = yaml.safe_load(file_content)
    return yaml_data


async def yaml_dump(data: Any, file: Union[Path, str]) -> Optional[Union[Path, str]]:
    """
    Write YAML to a file

    :param data: any data compatible with yaml.safe_dump
    :param file: a file path to write
    :return: if the dump succeeds, return the file, or None
    """
    dump = yaml.safe_dump(data)
    async with aiofiles.open(file, mode="w") as dst:
        await dst.write(dump)
    file_path = Path(file)
    if file_path.is_file() and file_path.stat().st_size > 0:
        LOGGER.info("Saved YAML to %s", file)
        return file
    LOGGER.error("Failed to save YAML to %s", file)



from typing import Optional

from botocore.exceptions import ClientError

from aio_aws.logger import get_logger

LOGGER = get_logger(__name__)


def response_code(response) -> int:
    return int(response.get("ResponseMetadata", {}).get("HTTPStatusCode"))


def response_success(response) -> bool:
    code = response_code(response)
    if code:
        return 200 <= code < 300
    else:
        return False


def handle_head_error_code(error: ClientError, item: str = None) -> Optional[bool]:
    err_code = error.response.get("Error", {}).get("Code")
    if err_code == "401":
        LOGGER.debug("GET HEAD 401: invalid credentials: %s", item)
        return False
    if err_code == "403":
        LOGGER.debug("GET HEAD 403: permission denied: %s", item)
        return False
    if err_code == "404":
        LOGGER.debug("GET HEAD 404: object missing: %s", item)
        return False
    return None



# Copyright 2020 Darren Weber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Basic async main loop
"""

import asyncio
from typing import Awaitable


def main(run: Awaitable):
    """Run main loop"""
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    try:
        loop.run_until_complete(run)
    finally:
        loop.stop()
        loop.close()



#! /usr/bin/env python3
# pylint: disable=bad-continuation

# Copyright 2020 Darren Weber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
AioAWS Batch
============

In testing this, it's able to run and monitor 100s of jobs from a laptop with modest
CPU/RAM resources.  It can recover from a crash by using a db-state, without re-running
the same jobs (by jobName).  It should be able to scale to run 1000s of jobs.

To run the example, the ``aio_aws.aio_aws_batch`` module has a ``main`` that will run and
manage about 5 live batch jobs (very small ``sleep`` jobs that don't cost much to run).  The
job state is persisted to ``aws_batch_jobs.json`` and if it runs successfully, it will not
run the jobs again; the `TinyDB`_ is used to recover job state by ``jobName``.  (This demo
assumes that some simple AWS Batch infrastructure exists already.)

.. code-block::

    # setup the python virtualenv
    # check the main details and modify for a preferred batch queue/CE and AWS region

    $ ./aio_aws/aio_aws_batch.py

    Test async batch jobs

    # wait a few minutes and watch the status messages
    # it submits and monitors the jobs until they complete
    # job status is saved and updated in `aws_batch_jobs.json`
    # when it's done, run it again and see that nothing is re-submitted

    $ ./aio_aws/aio_aws_batch.py

If the job monitoring is halted for some reason (like ``CNT-C``), it can recover from
the db-state, e.g.

.. code-block::

    $ ./aio_aws/aio_aws_batch.py

    Test async batch jobs
    [INFO]  2020-03-05T14:51:53.372Z  aio-aws:<module>:485  AWS Batch job (test-sleep-job-0000) recovered from db
    [INFO]  2020-03-05T14:51:53.373Z  aio-aws:<module>:485  AWS Batch job (test-sleep-job-0001) recovered from db
    [INFO]  2020-03-05T14:51:53.373Z  aio-aws:<module>:485  AWS Batch job (test-sleep-job-0002) recovered from db
    [INFO]  2020-03-05T14:51:53.374Z  aio-aws:<module>:485  AWS Batch job (test-sleep-job-0003) recovered from db
    [INFO]  2020-03-05T14:51:53.374Z  aio-aws:<module>:485  AWS Batch job (test-sleep-job-0004) recovered from db
    [INFO]  2020-03-05T14:51:53.690Z  aio-aws:aio_batch_job_waiter:375  AWS Batch job (846d54d4-c3c3-4a3b-9101-646d78d3bbfb) status: RUNNABLE
    [INFO]  2020-03-05T14:51:53.692Z  aio-aws:aio_batch_job_waiter:375  AWS Batch job (dfce3461-9eab-4f5b-846c-6f223d593f6f) status: RUNNABLE
    [INFO]  2020-03-05T14:51:53.693Z  aio-aws:aio_batch_job_waiter:375  AWS Batch job (637e6b27-8d4d-4f45-b988-c00775461616) status: RUNNABLE
    [INFO]  2020-03-05T14:51:53.701Z  aio-aws:aio_batch_job_waiter:375  AWS Batch job (d9ac27c9-e7d3-49cd-8f53-c84a9b4c1750) status: RUNNABLE
    [INFO]  2020-03-05T14:51:53.732Z  aio-aws:aio_batch_job_waiter:375  AWS Batch job (7ebfe7c4-44a4-40d6-9eab-3708e334689d) status: RUNNABLE

The batch data is a `TinyDB`_ json file, e.g.

.. code-block::

    >>> import json
    >>> jobs_db_file = '/tmp/aio_batch_jobs.json'
    >>> logs_db_file = '/tmp/aio_batch_logs.json'

    >>> with open(jobs_db_file) as job_file:
    ...     batch_data = json.load(job_file)
    ...
    >>> len(batch_data['aws-batch-jobs'])
    5

    >>> import tinydb
    >>> tinydb.TinyDB.DEFAULT_TABLE = "aws-batch-jobs"
    >>> tinydb.TinyDB.DEFAULT_TABLE_KWARGS = {"cache_size": 0}
    >>> # the jobs and logs are kept in separate DBs so
    >>> # that performance on the jobs_db is not compromised
    >>> # by too much data from job logs.
    >>> jobs_db = tinydb.TinyDB(jobs_db_file)
    >>> logs_db = tinydb.TinyDB(logs_db_file)


For the demo to run quickly, most of the module settings are fit for fast jobs.  For
much longer running jobs, there are functions that only submit jobs or check jobs and
the settings should be changed for monitoring jobs to only check every 10 or 20 minutes.

Monitoring Jobs
***************

The :py:func:`aio_aws.aio_aws_batch.aio_batch_job_manager` can submit a job, wait
for it to complete and retry if it fails on a SPOT termination. It saves the job status
using the :py:class:`aio_aws.aio_aws_batch import AWSBatchDB`.  The job manager
uses :py:func:`aio_aws.aio_aws_batch.aio_batch_job_waiter`, which uses these settings
to control the async-wait between polling the job status:

- :py:const:`aio_aws.aio_aws_config.BATCH_STARTUP_PAUSE`
- :py:const:`aio_aws.aio_aws_config.MAX_PAUSE`
- :py:const:`aio_aws.aio_aws_config.MIN_PAUSE`

These settings control how often job descriptions are polled.  These requests for job status
are also limited by the client connection pool and the client semaphore used by the job
manager.  Since AWS Batch has API limits on the number of requests for job status, it's best
to use a client connection pool and semaphore of about 10 connections.  Any failures to poll
for a job status will be retried a few times (using some random jitter on retry rates).

To modify the polling frequency settings, use a custom config.  For example, the unit
test suite uses much faster polling on mock batch jobs to speed up the unit tests; e.g.

.. code-block::

    config = AWSBatchConfig(
        start_pause=0.4, min_pause=0.8, max_pause=1.0, min_jitter=0.1, max_jitter=0.2,
    )


Getting Started
***************

The example above uses code similar to the following.  The following code appears
to work like regular synchronous code; behind the scenes, it wraps asyncio coroutines
and event loops.

To fully understand it, look at the module code directly.  For custom solutions, use
an event loop and some of the coroutines available.  Note that using
asyncio for AWS services requires the `aiobotocore`_ library, which wraps a
release of `botocore`_ to patch it with features for async coroutines using
`asyncio`_ and `aiohttp`_.  To avoid issuing too many concurrent requests (DOS attack),
the async approach should use a client connection limiter, based on ``asyncio.Semaphore()``.
It's recommended to use a single session and a single client with a connection pool.
Although there are context manager patterns, it's also possible to manage closing the client
after everything is done.

.. code-block::

    # python 3.6

    import asyncio

    from aio_aws.aio_aws_batch import AWSBatchConfig
    from aio_aws.aio_aws_batch import AWSBatchDB
    from aio_aws.aio_aws_batch import AWSBatchJob
    from aio_aws.aio_aws_batch import jobs_recovery
    from aio_aws.aio_aws_batch import jobs_to_run
    from aio_aws.aio_aws_batch import aio_batch_run_jobs
    from aio_aws.aio_aws_batch import aio_batch_get_logs

    # For the `aws_region` where batch will run:
    # - create a `batch-dev` compute environment
    # - create a `batch-dev` batch queue
    # - create a `batch-dev` job definition using alpine:latest

    job_command = "for a in `seq 1 10`; do echo Hello $a; sleep 0.5; done"

    # Create a list of AWSBatchJob objects, which are simple wrappers
    # on AWS Batch job data.  These objects are simple, by design, to
    # allow them to be used in either synchronous or asynchronous code.
    batch_jobs = []
    for i in range(20):
        job_name = f"test-logs-job-{i:04d}"

        # Creating an AWSBatchJob instance does not run anything, it's simply
        # a dataclass to retain and track job attributes.
        # Replace 'command' with 'container_overrides' dict for more options;
        # do not use 'command' together with 'container_overrides'.
        batch_job = AWSBatchJob(
            job_name=job_name,
            job_definition="batch-dev",
            job_queue="batch-dev",
            command=["/bin/sh", "-c", job_command],
        )
        batch_jobs.append(batch_job)

    # The AWSBatchDB wraps TinyDB data stores; it can be
    # used in either synchronous or asynchronous code.
    batch_jobs_db = AWSBatchDB(
        jobs_db_file="/tmp/aio_batch_jobs.json",
        logs_db_file="/tmp/aio_batch_logs.json",
    )

    batch_jobs = jobs_recovery(jobs=batch_jobs, jobs_db=batch_jobs_db)
    batch_jobs = jobs_to_run(jobs=batch_jobs, jobs_db=batch_jobs_db)
    if batch_jobs:

        # Create an event loop for the aio processing
        loop = asyncio.get_event_loop()
        loop.set_debug(enabled=True)
        try:

            # for polling frequency of 10-30 seconds, with 30-60 second job starts;
            # use lower frequency polling for long running jobs
            aio_config = AWSBatchConfig(
                max_pool_connections=10,
                aws_region="us-west-2",
                batch_db=batch_jobs_db,
                min_pause=10,
                max_pause=30,
                start_pause=30,
            )

            loop.run_until_complete(aio_batch_run_jobs(jobs=batch_jobs, config=aio_config))

            complete_jobs = []
            for job in batch_jobs:
                if job.status in ["SUCCEEDED", "FAILED"]:
                    complete_jobs.append(job)

            loop.run_until_complete(aio_batch_get_logs(jobs=complete_jobs, config=aio_config))

        finally:
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.stop()
            loop.close()


.. seealso::
    - https://aiobotocore.readthedocs.io/en/latest/
    - https://botocore.amazonaws.com/v1/documentation/api/latest/index.html

"""

import asyncio
import os
import re
from asyncio import AbstractEventLoop
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional

import aiobotocore.client  # type: ignore
import aiobotocore.session  # type: ignore
import botocore.endpoint  # type: ignore
import botocore.exceptions  # type: ignore
import botocore.session  # type: ignore
import tinydb

from aio_aws.aio_aws_config import AioAWSConfig
from aio_aws.aio_aws_config import delay
from aio_aws.aio_aws_config import jitter
from aio_aws.utils import response_success
from aio_aws.logger import get_logger

LOGGER = get_logger(__name__)

#: batch job startup pause (seconds)
BATCH_STARTUP_PAUSE: float = 30


@dataclass
class AWSBatchJobDescription:
    jobName: str = None
    jobId: str = None
    jobQueue: str = None
    status: str = None
    attempts: List[Dict] = None
    statusReason: str = None
    createdAt: int = None
    startedAt: int = None
    stoppedAt: int = None
    dependsOn: List[str] = None
    jobDefinition: str = None
    parameters: Dict = None
    container: Dict = None
    timeout: Dict = None


@dataclass
class AWSBatchJob:
    """
    AWS Batch job

    Creating an AWSBatchJob instance does not run anything, it's simply
    a dataclass to retain and track job attributes.  There are no instance
    methods to run a job, the instances are passed to async coroutine functions.

    Replace 'command' with 'container_overrides' dict for more options;
    do not use 'command' together with 'container_overrides', or understand
    that the construction will update the `container_overrides['command']`
    when the `command is not None`.

    :param job_name: A job jobName (truncated to 128 characters).
    :param job_definition: A job job_definition.
    :param job_queue: A batch queue.
    :param command: A container command.
    :param depends_on: list of dictionaries like:

        .. code-block::

            [
              {'jobId': 'abc123', ['type': 'N_TO_N' | 'SEQUENTIAL'] },
            ]

        type is optional, used only for job arrays

    :param container_overrides: a dictionary of container overrides.
        Overrides include 'vcpus', 'memory', 'instanceType',
        'environment', and 'resourceRequirements'. If the `command`
        parameter is defined, it overrides the `container_overrides['command']`

    :param max_tries: an optional limit to the number of job retries, which
        can apply in the job-manager function to any job with a SPOT failure
        only and it applies regardless of the job-definition settings.

    .. seealso::
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html
    """

    STATUSES = [
        "SUBMITTED",
        "PENDING",
        "RUNNABLE",
        "STARTING",
        "RUNNING",
        "SUCCEEDED",
        "FAILED",
    ]

    job_name: str
    job_queue: str
    job_definition: str
    command: List[str] = None
    depends_on: List[Dict] = None
    container_overrides: Dict = None
    job_id: Optional[str] = None
    status: Optional[str] = None
    job_tries: List[str] = None
    num_tries: int = 0
    max_tries: int = 4
    job_submission: Optional[Dict] = None
    job_description: Optional[Dict] = None
    logs: Optional[List[Dict]] = None

    def __post_init__(self):

        self.job_name = self.job_name[:128]

        if self.job_tries is None:
            self.job_tries = []

        if self.depends_on is None:
            self.depends_on = []

        if self.container_overrides is None:
            self.container_overrides = {}

        if self.command:
            self.container_overrides.update({"command": self.command})

    @property
    def params(self):
        """AWS Batch parameters for job submission"""
        return {
            "jobName": self.job_name,
            "jobQueue": self.job_queue,
            "jobDefinition": self.job_definition,
            "containerOverrides": self.container_overrides,
            "dependsOn": self.depends_on,
        }

    @property
    def db_data(self) -> Dict:
        """AWS Batch job data for state machine persistence"""
        # The job.logs are NOT included here, by design.
        return {
            "job_id": self.job_id,
            "job_name": self.job_name,
            "job_queue": self.job_queue,
            "job_definition": self.job_definition,
            "job_submission": self.job_submission,
            "job_description": self.job_description,
            "container_overrides": self.container_overrides,
            "command": self.command,
            "depends_on": self.depends_on,
            "status": self.status,
            "job_tries": self.job_tries,
            "num_tries": self.num_tries,
            "max_tries": self.max_tries,
        }

    @property
    def db_logs_data(self) -> Optional[Dict]:
        """AWS Batch job with logs persistence"""
        if self.logs:
            data = self.db_data
            data["logs"] = self.logs
            return data

    def reset(self):
        """Clear the job_id and all related job data"""
        self.job_id = None
        self.job_description = None
        self.job_submission = None
        self.status = None
        self.logs = None

    @property
    def created(self) -> Optional[int]:
        if self.job_description:
            return self.job_description.get("createdAt")

    @property
    def started(self) -> Optional[int]:
        if self.job_description:
            return self.job_description.get("startedAt")

    @property
    def stopped(self) -> Optional[int]:
        if self.job_description:
            return self.job_description.get("stoppedAt")

    @property
    def elapsed(self) -> Optional[int]:
        created = self.created
        stopped = self.stopped
        if stopped and created:
            return stopped - created

    @property
    def runtime(self) -> Optional[int]:
        started = self.started
        stopped = self.stopped
        if started and stopped:
            return stopped - started

    @property
    def spinup(self) -> Optional[int]:
        created = self.created
        started = self.started
        if started and created:
            return started - created


@dataclass
class AWSBatchDB:
    """
    AWS Batch job databases

    The jobs and logs are kept in separate DBs so
    that performance on the jobs_db is not compromised
    by too much data from job logs.  This makes it possible
    to use the jobs_db without capturing logs as well.

    .. seealso:: https://tinydb.readthedocs.io/en/latest/
    """

    #: a file used for :py:class::`TinyDB(jobs_db_file)`
    jobs_db_file: str = "/tmp/aws_batch_jobs.json"

    #: a file used for :py:class::`TinyDB(logs_db_file)`
    logs_db_file: str = "/tmp/aws_batch_logs.json"

    def __post_init__(self):

        tinydb.TinyDB.DEFAULT_TABLE = "aws-batch-jobs"
        tinydb.TinyDB.DEFAULT_TABLE_KWARGS = {"cache_size": 0}

        # the jobs and logs are kept in separate DBs so
        # that performance on the jobs_db is not compromised
        # by too much data from job logs.
        self.jobs_db = tinydb.TinyDB(self.jobs_db_file)
        self.logs_db = tinydb.TinyDB(self.logs_db_file)
        LOGGER.info("Using batch-jobs-db file: %s", self.jobs_db_file)
        LOGGER.info("Using batch-logs-db file: %s", self.logs_db_file)

        # Lazy init for any asyncio instances
        self._db_sem = None

    @property
    def db_semaphore(self) -> asyncio.Semaphore:
        """A semaphore to limit requests to the db"""
        if self._db_sem is None:
            self._db_sem = asyncio.Semaphore()
        return self._db_sem

    def find_by_job_id(self, job_id: str) -> Optional[tinydb.database.Document]:
        """
        Find one job by the jobId

        :param job_id: a batch jobId
        :return: the :py:meth:`AWSBatchJob.job_data` or None
        """
        if job_id:
            job_query = tinydb.Query()
            db_result = self.jobs_db.get(job_query.job_id == job_id)
            if db_result:
                return db_result

    def find_by_job_name(self, job_name: str) -> List[tinydb.database.Document]:
        """
        Find any jobs matching the jobName

        :param job_name: a batch jobName
        :return: a list of documents containing :py:meth:`AWSBatchJob.job_data`
        """
        if job_name:
            job_query = tinydb.Query()
            return self.jobs_db.search(job_query.job_name == job_name)

    def find_latest_job_name(self, job_name: str) -> AWSBatchJob:
        """
        Find the latest job matching the jobName,
        based on the "createdAt" time stamp.

        :param job_name: a batch jobName
        :return: the latest job record available
        """
        jobs_saved = self.find_by_job_name(job_name)
        if jobs_saved:
            db_jobs = [AWSBatchJob(**job_doc) for job_doc in jobs_saved]
            db_jobs = sorted(db_jobs, key=lambda j: j.created)
            db_job = db_jobs[-1]
            return db_job

    def remove_by_job_id(self, job_id: str) -> Optional[tinydb.database.Document]:
        """
        Remove any job matching the jobId

        :param job_id: a batch jobId
        :return: a deleted document
        """
        if job_id:
            job = self.find_by_job_id(job_id)
            if job:
                self.jobs_db.remove(doc_ids=[job.doc_id])
                return job

    def remove_by_job_name(self, job_name: str) -> List[tinydb.database.Document]:
        """
        Remove any jobs matching the jobName

        :param job_name: a batch jobName
        :return: a list of deleted documents
        """
        if job_name:
            jobs_found = self.find_by_job_name(job_name)
            if jobs_found:
                docs = [doc.doc_id for doc in jobs_found]
                self.jobs_db.remove(doc_ids=docs)
            return jobs_found

    def save_job(self, job: AWSBatchJob) -> List[int]:
        """
        Insert or update a job (if it has a job_id)

        :param job: an AWSBatchJob
        :return: a List[tinydb.database.Document.doc_id]
        """
        if job.job_id:
            job_query = tinydb.Query()
            return self.jobs_db.upsert(job.db_data, job_query.job_id == job.job_id)
        else:
            LOGGER.error("FAIL to save_job without job_id")

    def find_job_logs(self, job_id: str) -> Optional[tinydb.database.Document]:
        """
        Find job logs by job_id

        :param job_id: str
        :return: a tinydb.database.Document with job logs
        """
        if job_id:
            log_query = tinydb.Query()
            db_result = self.logs_db.get(log_query.job_id == job_id)
            if db_result:
                return db_result
        else:
            LOGGER.error("FAIL to find_job_logs without job_id")

    def save_job_logs(self, job: AWSBatchJob) -> List[int]:
        """
        Insert or update job logs (if the job has a job_id)

        :param job: an AWSBatchJob
        :return: a List[tinydb.database.Document.doc_id]
        """
        if job.job_id:
            log_query = tinydb.Query()
            return self.logs_db.upsert(job.db_logs_data, log_query.job_id == job.job_id)
        else:
            LOGGER.error("FAIL to save_job_logs without job_id")

    def find_jobs_to_run(self) -> List[AWSBatchJob]:
        """
        Find all jobs that have not SUCCEEDED.  Note that any jobs handled
        by the job-manager will not re-run if they have a job.job_id, those
        jobs will be monitored until complete.
        """
        jobs = [AWSBatchJob(**job_doc) for job_doc in self.jobs_db.all()]
        jobs_outstanding = []
        for job in jobs:
            LOGGER.info(
                "AWS Batch job (%s:%s) has db status: %s",
                job.job_name,
                job.job_id,
                job.status,
            )
            if job.job_id and job.status == "SUCCEEDED":
                LOGGER.debug(job.job_description)
                continue

            jobs_outstanding.append(job)

        return jobs_outstanding


def jobs_to_run(jobs: List[AWSBatchJob], jobs_db: AWSBatchDB) -> List[AWSBatchJob]:
    """
    Use the job.job_name to find any jobs_db records and check the job status
    on the latest submission of any job matching the job-name.  For any job that
    has a matching job-name in the jobs_db, check whether the job has already run and
    SUCCEEDED.  Any jobs that have already run and SUCCEEDED are not returned.

    Return all jobs that have not run yet or have not SUCCEEDED.  Note that any
    jobs subsequently input to the job-manager will not re-run if they already
    have a job.job_id, those jobs will be monitored until complete.  To re-run
    jobs that are returned from this filter function, first use `job.reset()`
    before passing the jobs to the job-manager.

    .. note::
        The input jobs are not modified in any way in this function, i.e. there
        are no updates from the jobs_db applied to the input jobs.

        To recover jobs from the jobs_db, use
        - :py:meth:`AWSBatchDB.find_jobs_to_run()`
        - :py:meth:`AWSBatchDB.jobs_db.all()`
    """
    jobs_outstanding = []
    for job in jobs:
        # Avoid side-effects in this function:
        # - treat the job as a read-only object
        # - treat any jobs_db records as read-only objects
        # - if a job is not saved, don't save it to jobs_db

        # First check the job itself before checking the jobs_db
        if job.job_id and job.status == "SUCCEEDED":
            LOGGER.debug(job.job_description)
            continue

        db_job = jobs_db.find_latest_job_name(job.job_name)
        if db_job:
            LOGGER.info(
                "AWS Batch job (%s:%s) has db status: %s",
                db_job.job_name,
                db_job.job_id,
                db_job.status,
            )

            if db_job.job_id and db_job.status == "SUCCEEDED":
                LOGGER.debug(db_job.job_description)
                continue

        jobs_outstanding.append(job)

    return jobs_outstanding


def jobs_recovery(jobs: List[AWSBatchJob], jobs_db: AWSBatchDB) -> List[AWSBatchJob]:
    """
    Use the job.job_name to find any jobs_db records to recover job data.
    """
    jobs_recovered = []
    for job in jobs:

        db_job = jobs_db.find_latest_job_name(job.job_name)
        if db_job:
            LOGGER.info(
                "AWS Batch job (%s:%s) recovered from db with status: %s",
                db_job.job_name,
                db_job.job_id,
                db_job.status,
            )
            jobs_recovered.append(db_job)
        else:
            jobs_recovered.append(job)

    return jobs_recovered


@dataclass
class AWSBatchConfig(AioAWSConfig):
    #: a batch job startup pause, ``random.uniform(start_pause, start_pause * 2)``;
    #: this applies when the job status is in ["SUBMITTED", "PENDING", "RUNNABLE"]
    start_pause: float = BATCH_STARTUP_PAUSE
    #: an optional AWSBatchDB
    batch_db: Optional[AWSBatchDB] = None

    def get_batch_db(self) -> AWSBatchDB:
        """
        Return the provided batch_db or create and return a default one

        :return: an AWSBatchDB
        """
        if self.batch_db is None:
            self.batch_db = AWSBatchDB()
        return self.batch_db

    @asynccontextmanager
    async def create_batch_client(self) -> aiobotocore.client.AioBaseClient:
        """
        Create and yield an AWS Batch client using the ``AWSBatchConfig.session``

        :yield: an aiobotocore.client.AioBaseClient for AWS Batch
        """
        async with self.session.create_client("batch") as client:
            yield client

    @asynccontextmanager
    async def create_logs_client(self) -> aiobotocore.client.AioBaseClient:
        """
        Create and yield an AWS CloudWatchLogs client using the ``AWSBatchConfig.session``

        :yield: an aiobotocore.client.AioBaseClient for AWS CloudWatchLogs
        """
        async with self.session.create_client("logs") as client:
            yield client


def parse_job_description(job_id: str, jobs: Dict) -> Optional[Dict]:
    """
    Extract a job description for ``job_id`` from ``jobs``
    :param job_id: an AWS Batch ``jobId``
    :param jobs: a response to AWS Batch job descriptions
    :return: a job description for ``job_id``
    """
    job_desc = None
    for job in jobs["jobs"]:
        if job["jobId"] == job_id:
            job_desc = job
    return job_desc


async def aio_batch_job_submit(job: AWSBatchJob, config: AWSBatchConfig = None) -> Dict:
    """
    Asynchronous coroutine to submit a batch job; for a successful
    job submission, the jobId and other submission details are
    updated on the job object and persisted to ``config.batch_db``.

    :param job: A set of job parameters
    :param config: settings for task pauses between retries
    :return: a job response
    :raises: botocore.exceptions.ClientError
    """
    # async with config.semaphore:
    #     async with config.session.create_client("batch") as batch_client:

    if config is None:
        config = AWSBatchConfig.get_default_config()

    async with config.create_batch_client() as batch_client:
        tries = 0
        while tries <= config.retries:
            tries += 1
            try:
                params = job.params
                LOGGER.debug("AWS Batch job params: %s", params)
                response = await batch_client.submit_job(**params)
                LOGGER.debug("AWS Batch job response: %s", response)

                if response_success(response):
                    job.job_id = response["jobId"]
                    job.job_submission = response
                    job.job_tries.append(job.job_id)
                    job.num_tries += 1
                    jobs_db = config.get_batch_db()
                    async with jobs_db.db_semaphore:
                        jobs_db.save_job(job)
                    LOGGER.info(
                        "AWS Batch job (%s:%s) submitted try: %d of %d",
                        job.job_name,
                        job.job_id,
                        job.num_tries,
                        job.max_tries,
                    )
                else:
                    # TODO: are there some submission failures that could be recovered here?
                    LOGGER.error(
                        "AWS Batch job (%s:xxx) submission failure.", job.job_name
                    )
                return response

            except botocore.exceptions.ClientError as err:
                error = err.response.get("Error", {})
                if error.get("Code") == "TooManyRequestsException":
                    if tries <= config.retries:
                        # add an extra random sleep period to avoid API throttle
                        await jitter(
                            "batch-job-submit", config.min_jitter, config.max_jitter
                        )
                    continue  # allow it to retry, if possible
                else:
                    raise
        else:
            raise RuntimeError("AWS Batch job submission exceeded retries")


async def aio_batch_job_status(
    jobs: List[str], config: AWSBatchConfig = None
) -> Optional[Dict]:
    """
    Asynchronous coroutine to issue a batch job description request

    :param jobs: a list of batch jobId
    :param config: settings for task pauses between retries
    :return: a describe_jobs response
    :raises: botocore.exceptions.ClientError
    """
    if config is None:
        config = AWSBatchConfig.get_default_config()

    async with config.create_batch_client() as batch_client:
        tries = 0
        while tries <= config.retries:
            tries += 1
            try:
                return await batch_client.describe_jobs(jobs=jobs)
            except botocore.exceptions.ClientError as err:
                error = err.response.get("Error", {})
                if error.get("Code") == "TooManyRequestsException":
                    if tries <= config.retries:
                        # add an extra random sleep period to avoid API throttle
                        await jitter(
                            "batch-job-status", config.min_jitter, config.max_jitter
                        )
                    continue  # allow it to retry, if possible
                else:
                    raise
        else:
            raise RuntimeError("AWS Batch job description exceeded retries")


async def aio_batch_job_logs(
    job: AWSBatchJob, config: AWSBatchConfig = None
) -> Optional[List[Dict]]:
    """
    Asynchronous coroutine to get logs for a batch job log stream.  All
    the events available for a log stream are collected and returned.  The
    AWS Cloud Watch log streams can have delays, don't expect near real-time
    values in these data.

    :param job: A set of job parameters
    :param config: settings for task pauses between retries
    :return: a list of all the log events from get_log_events responses
    :raises: botocore.exceptions.ClientError

    .. seealso::
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/services/logs.html
    """
    if config is None:
        config = AWSBatchConfig.get_default_config()

    if not job.job_description:
        LOGGER.warning("AWS Batch job has no description, cannot fetch logs")
        return

    log_stream_name = job.job_description.get("container", {}).get("logStreamName")
    if not log_stream_name:
        LOGGER.warning(
            "AWS Batch job (%s:%s) has no log stream.", job.job_name, job.job_id
        )
        return

    async with config.create_logs_client() as logs_client:
        tries = 0
        while tries <= config.retries:
            tries += 1
            try:
                log_events = []

                forward_token = None
                while True:
                    kwargs = {
                        "logGroupName": "/aws/batch/job",
                        "logStreamName": log_stream_name,
                        "startFromHead": True,
                        # "startTime": 123,
                        # "endTime": 123,
                        # "limit": 123,
                    }
                    if forward_token:
                        kwargs["nextToken"] = forward_token

                    response = await logs_client.get_log_events(**kwargs)

                    events = response.get("events", [])
                    log_events.extend(events)

                    if forward_token != response["nextForwardToken"]:
                        forward_token = response["nextForwardToken"]
                    else:
                        break

                if log_events:
                    LOGGER.info(
                        "AWS Batch job (%s:%s) log events: %d",
                        job.job_name,
                        job.job_id,
                        len(log_events),
                    )
                    job.logs = log_events
                    jobs_db = config.get_batch_db()
                    async with jobs_db.db_semaphore:
                        jobs_db.save_job_logs(job)
                else:
                    LOGGER.warning(
                        "AWS Batch job (%s:%s) has no log events",
                        job.job_name,
                        job.job_id,
                    )

                return log_events

            except botocore.exceptions.ClientError as err:
                error = err.response.get("Error", {})
                if error.get("Code") == "TooManyRequestsException":
                    if tries <= config.retries:
                        # add an extra random sleep period to avoid API throttle
                        await jitter(
                            "batch-job-logs", config.min_jitter, config.max_jitter
                        )
                    continue  # allow it to retry, if possible
                else:
                    raise
        else:
            raise RuntimeError("AWS Batch job logs exceeded retries")


async def aio_batch_job_terminate(
    job_id: str, reason: str, config: AWSBatchConfig = None
) -> Dict:
    """
    Asynchronous coroutine to terminate a batch job

    :param job_id: a batch jobId
    :param reason: a reason to terminate the job
    :param config: settings for task pauses between retries
    :return: a job response
    :raises: botocore.exceptions.ClientError
    """
    if config is None:
        config = AWSBatchConfig.get_default_config()

    async with config.create_batch_client() as batch_client:

        tries = 0
        while tries <= config.retries:
            tries += 1
            try:
                LOGGER.info("AWS Batch job to terminate: %s, %s", job_id, reason)
                response = await batch_client.terminate_job(jobId=job_id, reason=reason)
                LOGGER.info("AWS Batch job response: %s", response)
                return response

            except botocore.exceptions.ClientError as err:
                error = err.response.get("Error", {})
                if error.get("Code") == "TooManyRequestsException":
                    if tries <= config.retries:
                        # add an extra random sleep period to avoid API throttle
                        await jitter(
                            "batch-job-terminate", config.min_jitter, config.max_jitter
                        )
                    continue  # allow it to retry, if possible
                else:
                    raise
        else:
            raise RuntimeError("AWS Batch job termination exceeded retries")


async def aio_batch_job_waiter(
    job: AWSBatchJob, config: AWSBatchConfig = None
) -> Optional[Dict]:
    """
    Asynchronous coroutine to wait on a batch job.  There is no explict
    timeout on a job waiter, it depends on setting a timeout on the batch
    job definition.  The job waiter exits when the batch job status is
    either "SUCCEEDED" or "FAILED".  The job data is updated each time
    the job description is polled (it's also saved to the jobs-db).

    job status identifiers are assumed to be:
    ["SUBMITTED", "PENDING", "RUNNABLE", "STARTING", "RUNNING", "FAILED", "SUCCEEDED"]

    :param job: a batch job
    :param config: settings for task pauses between retries
    :return: a describe_jobs response for job.job_id when it's complete
    :raises: botocore.exceptions.ClientError
    """
    if config is None:
        config = AWSBatchConfig.get_default_config()

    try:
        monitor_failures = 0
        while True:
            response = await aio_batch_job_status([job.job_id], config)
            LOGGER.debug(
                "AWS Batch job (%s:%s) status: %s", job.job_name, job.job_id, response
            )

            job_desc = parse_job_description(job.job_id, response)
            if job_desc:

                job.job_description = job_desc
                job.status = job_desc["status"]
                jobs_db = config.get_batch_db()
                async with jobs_db.db_semaphore:
                    jobs_db.save_job(job)  # TODO: use async-db

                LOGGER.info(
                    "AWS Batch job (%s:%s) status: %s",
                    job.job_name,
                    job.job_id,
                    job.status,
                )

                if job_desc["status"] in ["FAILED", "SUCCEEDED"]:
                    return job_desc

                if job_desc["status"] in ["SUBMITTED", "PENDING", "RUNNABLE"]:
                    # Wait longer than regular pause to allow job startup
                    await delay(
                        job.job_name, config.start_pause, config.start_pause * 2
                    )

            else:
                monitor_failures += 1
                if monitor_failures <= config.retries:
                    LOGGER.warning(
                        "AWS Batch job (%s:%s) retry job description",
                        job.job_name,
                        job.job_id,
                    )
                else:
                    LOGGER.error(
                        "AWS Batch job (%s:%s) failed to monitor job",
                        job.job_name,
                        job.job_id,
                    )
                    break

            # Allow the job status to be checked first, in case it is complete.
            await delay(job.job_name, config.min_pause, config.max_pause)

    except botocore.exceptions.ClientError as err:
        LOGGER.error(
            "AWS Batch job (%s:%s) failed to monitor job", job.job_name, job.job_id
        )
        raise


async def aio_batch_job_manager(
    job: AWSBatchJob, config: AWSBatchConfig = None
) -> Optional[Dict]:
    """
    Asynchronous coroutine to manage a batch job.  The job-manager
    will update attributes on the job as the job passes through
    submission and completion.  On completion, the job logs are
    retrieved.  As the job status and logs are collected, the
    job data is persisted to the jobs_db.

    Note that any job with a job.job_id will not re-run, those
    jobs will be monitored until complete.  To re-run a job that
    has already run, first call the job.reset() method to clear
    any previous job state.  (Any previous attempts are recorded
    in job.job_tries and in the job.job_description.)

    :param job: a batch job spec
    :param config: settings for task pauses between retries
    :return: a describe_jobs response for job_id when it's complete
    :raises: botocore.exceptions.ClientError
    """
    if config is None:
        config = AWSBatchConfig.get_default_config()

    while True:

        # jobs with an existing job.job_id skip submission to be waited
        if job.job_id is None:
            if job.num_tries < job.max_tries:
                # the job submission can update the job data and jobs-db
                response = await aio_batch_job_submit(job, config=config)
                if not response_success(response):
                    # If the job-submission has failed, don't retry it
                    # because there is probably something to fix.
                    return
            else:
                LOGGER.warning(
                    "AWS Batch job (%s) exceeds retries: %d of %d",
                    job.job_name,
                    job.num_tries,
                    job.max_tries,
                )
                return

        # the job waiter updates the job description and status
        await aio_batch_job_waiter(job, config=config)

        if job.status == "FAILED":
            # SPOT failure requires a retry, it usually has
            # statusReason: Host EC2 (instance {instanceId}) terminated
            #
            # Some common reasons with no retry behavior:
            # - user initiated job termination with a custom reason
            # - "Dependent Job failed" - unknown upstream failure
            # - "Essential container in task exited" - container error
            reason = job.job_description.get("statusReason", "")
            if re.match(r"Host EC2.*terminated", reason):
                LOGGER.warning(
                    "AWS Batch job (%s:%s) SPOT failure, run retry.",
                    job.job_name,
                    job.job_id,
                )
                job.reset()  # reset jobs are not saved to the jobs-db
                continue

        return job.job_description


async def handle_tasks_as_completed(tasks, loop=None):
    for task in asyncio.as_completed(tasks, loop=loop):
        job_desc = await task
        LOGGER.debug(job_desc)


async def aio_batch_run_jobs(
    jobs: List[AWSBatchJob], config: AWSBatchConfig, loop: AbstractEventLoop = None
):
    if loop is None:
        # Ensure that an event loop is running in this thread
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

    batch_tasks = [
        loop.create_task(aio_batch_job_manager(job=job, config=config)) for job in jobs
    ]
    await handle_tasks_as_completed(batch_tasks)

    # Using the asyncio.Task API
    for task in batch_tasks:
        if task.done():
            try:
                task.exception()
            except asyncio.CancelledError as err:
                LOGGER.exception(err)

            result = task.result()
            LOGGER.debug(result)


async def aio_batch_get_logs(
    jobs: List[AWSBatchJob], config: AWSBatchConfig, loop: AbstractEventLoop = None
):
    if loop is None:
        loop = asyncio.get_event_loop()

    batch_tasks = [
        loop.create_task(aio_batch_job_logs(job=job, config=config)) for job in jobs
    ]
    await handle_tasks_as_completed(batch_tasks)

    # Using the asyncio.Task API
    for task in batch_tasks:
        if task.done():
            try:
                task.exception()
            except asyncio.CancelledError as err:
                LOGGER.exception(err)

            result = task.result()
            LOGGER.debug(result)


if __name__ == "__main__":

    print()
    print("Test async batch jobs")

    # job_command = "for a in `seq 1 10000`; do echo Hello $a; done"
    job_command = "for a in `seq 1 10`; do echo Hello $a; sleep 0.5; done"

    n_jobs = 20
    batch_jobs = []
    for i in range(n_jobs):
        job_name = f"test-job-{i:04d}"

        # Creating an AWSBatchJob instance does not run anything, it's simply
        # a dataclass to retain and track job attributes.
        # Replace 'command' with 'container_overrides' dict for more options;
        # do not use 'command' together with 'container_overrides'.
        batch_job = AWSBatchJob(
            job_name=job_name,
            job_definition="batch-dev",
            job_queue="batch-dev",
            command=["/bin/sh", "-c", job_command],
        )
        batch_jobs.append(batch_job)

    batch_jobs_db = AWSBatchDB(
        jobs_db_file="/tmp/aio_batch_jobs.json", logs_db_file="/tmp/aio_batch_logs.json"
    )

    batch_jobs = jobs_recovery(jobs=batch_jobs, jobs_db=batch_jobs_db)
    batch_jobs = jobs_to_run(jobs=batch_jobs, jobs_db=batch_jobs_db)
    if batch_jobs:

        # Create an event loop for the aio processing
        loop = asyncio.get_event_loop()
        loop.set_debug(enabled=True)
        try:

            # for polling frequency of 5-20 seconds, with 30-60 second job starts
            aio_config = AWSBatchConfig(
                aws_region=os.getenv("AWS_DEFAULT_REGION", "us-west-2"),
                batch_db=batch_jobs_db,
                max_pool_connections=10,
                min_pause=5,
                max_pause=20,
                start_pause=30,
            )

            loop.run_until_complete(
                aio_batch_run_jobs(jobs=batch_jobs, config=aio_config)
            )

            complete_jobs = []
            for job in batch_jobs:
                if job.status in ["SUCCEEDED", "FAILED"]:
                    complete_jobs.append(job)

            loop.run_until_complete(
                aio_batch_get_logs(jobs=complete_jobs, config=aio_config)
            )

        finally:
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.stop()
            loop.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aio_aws']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.2,<6.0',
 'aiobotocore>=1.3.0,<2.0.0',
 'aiofiles>=0.7.0,<0.8.0',
 'boto3>=1.17.42,<1.18.0',
 'botocore>=1.20.42,<1.21.0',
 'requests>=2.23.0,<3.0.0',
 'tinydb>=3.15,<4.0']

extras_require = \
{'all': ['aioredis>=1.3,<2.0', 'databases[postgresql,mysql,sqlite]'],
 'awscli': ['awscli>=1.18,<2.0'],
 'docs': ['Sphinx>=3.2,<4.0',
          'sphinx-autoapi>=1.5,<2.0',
          'sphinx-autodoc-typehints>=1.11,<2.0',
          'sphinx-rtd-theme>=0.5.0,<0.6.0',
          'ipython>=7.24.1,<8.0.0']}

setup_kwargs = {
    'name': 'aio-aws',
    'version': '0.13.0',
    'description': 'aio-aws',
    'long_description': '[![Build Status](https://travis-ci.com/dazza-codes/aio-aws.svg?branch=master)](https://travis-ci.com/dazza-codes/aio-aws)\n[![Documentation Status](https://readthedocs.org/projects/aio-aws/badge/?version=latest)](https://aio-aws.readthedocs.io/en/latest/?badge=latest)\n\n# aio-aws\n\nAsynchronous functions and tools for AWS services.  There is a\nlimited focus on s3 and AWS Batch and Lambda.  Additional services could be\nadded, but this project is likely to retain a limited focus.\nFor general client solutions, see\n[aioboto3](https://github.com/terrycain/aioboto3) and\n[aiobotocore](https://github.com/aio-libs/aiobotocore), which wrap\n[botocore](https://botocore.amazonaws.com/v1/documentation/api/latest/index.html)\n\nThe API documentation is at [readthedocs](https://aio-aws.readthedocs.io/)\n\n# Install\n\nThis project has a very limited focus.  For general client solutions, see\n[aioboto3](https://github.com/terrycain/aioboto3) and\n[aiobotocore](https://github.com/aio-libs/aiobotocore), which wrap\n[botocore](https://botocore.amazonaws.com/v1/documentation/api/latest/index.html)\nto patch it with features for async coroutines using\n[aiohttp](https://aiohttp.readthedocs.io/en/latest/) and\n[asyncio](https://docs.python.org/3/library/asyncio.html).\n\nThis project is alpha-status with a 0.x.y API version that could break.\nThere is no promise to support or develop it extensively, at this time.\n\n## pip\n\n```shell\npip install -U aio-aws[all]\npip check  # pip might not guarantee consistent packages\n```\n\n## poetry\n\npoetry will try to guarantee consistent packages or fail.\n\n```shell\n# with optional extras\npoetry add aio-aws --extras all\n```\n\n```toml\n# pyproject.toml snippet\n\n[tool.poetry.dependencies]\npython = "^3.7"\n\n# with optional extras\naio-aws = {version = "^0.1.0", extras = ["all"]}\n\n# or, to make it an optional extra\naio-aws = {version = "^0.1.0", extras = ["all"], optional = true}\n[tool.poetry.extras]\naio-aws = ["aio-aws"]\n\n```\n\n# Contributing\n\nTo use the source code, it can be cloned directly. To\ncontribute to the project, first fork it and clone the forked repository.\n\nThe following setup assumes that\n[miniconda3](https://docs.conda.io/en/latest/miniconda.html) and\n[poetry](https://python-poetry.org/docs/) are installed already\n(and `make` 4.x).\n\n- https://docs.conda.io/en/latest/miniconda.html\n    - recommended for creating virtual environments with required versions of python\n    - see https://github.com/dazza-codes/conda_container/blob/master/conda_venv.sh\n- https://python-poetry.org/docs/\n    - recommended for managing a python project with pip dependencies for\n      both the project itself and development dependencies\n\n```shell\ngit clone https://github.com/dazza-codes/aio-aws\ncd aio-aws\nconda create -n aio-aws python=3.7\nconda activate aio-aws\nmake init  # calls poetry install\nmake test\n```\n\n# License\n\n```text\nCopyright 2019-2021 Darren Weber\n\nLicensed under the Apache License, Version 2.0 (the "License");\nyou may not use this file except in compliance with the License.\nYou may obtain a copy of the License at\n\n   http://www.apache.org/licenses/LICENSE-2.0\n\nUnless required by applicable law or agreed to in writing, software\ndistributed under the License is distributed on an "AS IS" BASIS,\nWITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\nSee the License for the specific language governing permissions and\nlimitations under the License.\n```\n\n# Notices\n\nInspiration for this project comes from various open source projects that use\nthe Apache 2 license, including but not limited to:\n- Apache Airflow: https://github.com/apache/airflow\n- aiobotocore: https://github.com/aio-libs/aiobotocore\n- botocore: https://github.com/boto/botocore\n',
    'author': 'Darren Weber',
    'author_email': 'dazza-codes@github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dazza-codes/aio-aws',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#! /usr/bin/env python3
# pylint: disable=bad-continuation

# Copyright 2020 Darren Weber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
AioAWS S3
=========

Example code for async AWS S3 services. The `aiobotocore`_ library wraps a
release of `botocore`_ to patch it with features for async coroutines using
`asyncio`_ and `aiohttp`_. The simple example code iterates on a list of
buckets or objects to issue a HEAD request, as a simple way to determine
whether access is permitted. The example compares an async approach (using
aiobotocore) vs a sequential blocking approach (using botocore). The async
approach uses a client connection limiter, based on ``asyncio.Semaphore(20)``.

.. code-block:: shell

    $ ./aio_aws/aio_aws_s3.py

    aio-aws check an s3 object that exists
    GET HEAD s3://noaa-goes16/index.html -> {'ResponseMetadata': {'RequestId': 'A2F4DF282C84341B', 'HostId': 'Rt45GH1taNRVUAPTg8XJfs4KLvri0OKos/6Wohp5tAfQa+moUT/9mC/0Wa9cYVQZcMAWKtIYfkE=', 'HTTPStatusCode': 200, 'HTTPHeaders': {'x-amz-id-2': 'Rt45GH1taNRVUAPTg8XJfs4KLvri0OKos/6Wohp5tAfQa+moUT/9mC/0Wa9cYVQZcMAWKtIYfkE=', 'x-amz-request-id': 'A2F4DF282C84341B', 'date': 'Tue, 03 Mar 2020 19:22:21 GMT', 'last-modified': 'Wed, 12 Feb 2020 19:14:27 GMT', 'etag': '"8f9810548f75fd8b8a16431c74ad54ac"', 'content-encoding': 'text/html', 'accept-ranges': 'bytes', 'content-type': 'text/html', 'content-length': '32356', 'server': 'AmazonS3'}, 'RetryAttempts': 0}, 'AcceptRanges': 'bytes', 'LastModified': datetime.datetime(2020, 2, 12, 19, 14, 27, tzinfo=tzutc()), 'ContentLength': 32356, 'ETag': '"8f9810548f75fd8b8a16431c74ad54ac"', 'ContentEncoding': 'text/html', 'ContentType': 'text/html', 'Metadata': {}}

    aio-aws check an s3 object that is missing; it raises.
    GET HEAD s3://noaa-goes16/missing.html -> 404: object missing

    aio-aws find all buckets that allow access.
    queried 72 buckets
    found 69 accessible buckets
    finished in 2.42 seconds.

    aio-aws collection of all objects in a bucket-prefix.
    found 9015 s3 objects
    finished in 4.43 seconds.

    aws-sequential find all buckets that allow access.
    queried 72 buckets
    found 69 accessible buckets
    finished in 31.11 seconds.

    aws-sequential collection for all objects in a bucket-prefix
    found 9015 s3 objects
    finished in 35.72 seconds.

    Checking equality of collections for aio-aws vs. aws-sync
    The collections for aio-aws vs. aws-sync match OK


Getting Started
***************

The example above uses code similar to the following.
Using asyncio for AWS services requires the `aiobotocore`_ library, which wraps a
release of `botocore`_ to patch it with features for async coroutines using
`asyncio`_ and `aiohttp`_.  To avoid issuing too many concurrent requests (DOS attack),
the async approach should use a client connection limiter, based on ``asyncio.Semaphore()``.
It's recommended to use a single session and a single client with a connection pool.
Although there are context manager patterns, it's also possible to manage closing the client
after everything is done.

.. code-block::

    # python 3.7

    import asyncio
    import aiobotocore.config
    import time

    from aio_aws.aio_aws_config import AioAWSConfig
    from aio_aws.aio_aws_s3 import aio_s3_objects_list

    # Use a single session with a larger connection pool.
    # The AioAWSConfig can provide clients on-demand with
    # a context manager pattern.
    aio_session = aio_aws_session()
    aio_config = AioAWSConfig(
        max_pool_connections=40,
        session=aio_session,
        min_pause=0.2,
        max_pause=0.6,
        min_jitter=0.2,
        max_jitter=0.8,
    )

    main_loop = asyncio.get_event_loop()

    try:
        # https://registry.opendata.aws/noaa-goes/
        noaa_goes_bucket = "noaa-goes16"
        noaa_prefix = "ABI-L2-ADPC/2019"  # use a prior year for stable results

        print("aio-aws collection of all objects in a bucket-prefix.")
        start = time.perf_counter()
        aio_s3_objects = main_loop.run_until_complete(
            aio_s3_objects_list(
                bucket_name=noaa_goes_bucket,
                bucket_prefix=noaa_prefix,
                config=aio_config
            )
        )
        aio_s3_uris = [f"s3://{noaa_goes_bucket}/{obj['Key']}" for obj in aio_s3_objects]
        print(f"found {len(aio_s3_uris)} s3 objects")
        end = time.perf_counter() - start
        print(f"finished in {end:0.2f} seconds.")

    finally:
        main_loop.run_until_complete(main_loop.shutdown_asyncgens())
        main_loop.stop()
        main_loop.close()


.. seealso::
    - https://aiobotocore.readthedocs.io/en/latest/
    - https://botocore.amazonaws.com/v1/documentation/api/latest/index.html
    - https://pawelmhm.github.io/asyncio/python/aiohttp/2016/04/22/asyncio-aiohttp.html
    - https://docs.python.org/3/library/asyncio.html
    - https://www.mathewmarcus.com/blog/asynchronous-aws-api-requests-with-asyncio.html

"""

import asyncio
import concurrent.futures
import time
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Tuple

from aiobotocore.client import AioBaseClient
from aiobotocore.config import AioConfig
from botocore import UNSIGNED
from botocore.client import BaseClient
from botocore.exceptions import ClientError

from aio_aws.aio_aws_config import aio_aws_session
from aio_aws.aio_aws_config import AioAWSConfig
from aio_aws.aio_aws_config import jitter
from aio_aws.utils import handle_head_error_code
from aio_aws.utils import response_success
from aio_aws.logger import get_logger

LOGGER = get_logger(__name__)


@dataclass(frozen=True)
class S3Parts:
    bucket: str
    key: str

    PROTOCOL: str = "s3://"
    DELIMITER: str = "/"

    @property
    def s3_uri(self):
        return f"{self.PROTOCOL}{self.bucket}{S3Parts.DELIMITER}{self.key}"

    @staticmethod
    def parse_s3_uri(uri: str) -> "S3Parts":
        uri = str(uri).strip()
        if not uri.startswith(S3Parts.PROTOCOL):
            raise ValueError(f"S3 URI must start with '{S3Parts.PROTOCOL}'")
        uri_path = uri.replace(S3Parts.PROTOCOL, "")
        uri_paths = uri_path.split(S3Parts.DELIMITER)
        bucket = uri_paths.pop(0)
        key = S3Parts.DELIMITER.join(uri_paths)

        return S3Parts(bucket, key)


def aws_s3_bucket_access(bucket_name: str, s3_client: BaseClient) -> bool:
    """
    Check access to an S3 bucket by issuing a HEAD request

    :param bucket_name: An s3 bucket name
    :param s3_client: botocore.client.BaseClient for s3
    :return: a tuple of ``(bucket_name: str, access: bool)``
    """
    try:
        head = s3_client.head_bucket(Bucket=bucket_name)
        if response_success(head):
            return True
        else:
            return False

    except ClientError as err:
        access = handle_head_error_code(err, bucket_name)
        if access is False:
            return False
        raise


def aws_s3_buckets_access(s3_client: BaseClient) -> Dict[str, bool]:
    """
    A concurrent.futures approach to list all s3 buckets
    and issue a HEAD request to check if access is allowed.  This
    is used in performance comparisons between botocore and
    aiobotocore functionality.

    :param s3_client: botocore.client.BaseClient for s3
    :return: a tuple of ``(bucket_name: str, access: bool)``
    """
    buckets = []
    response = s3_client.list_buckets()
    if response_success(response):
        for bucket in response["Buckets"]:
            bucket_name = bucket["Name"]
            buckets.append(bucket_name)

    # This pattern is FAST
    buckets_access = {}
    max_workers = s3_client.meta.config.max_pool_connections
    with concurrent.futures.ThreadPoolExecutor(max_workers) as executor:
        bucket_futures = {
            executor.submit(aws_s3_bucket_access, bucket, s3_client): bucket
            for bucket in buckets
        }
        for future in concurrent.futures.as_completed(bucket_futures):
            bucket = bucket_futures[future]
            access = future.result()
            buckets_access[bucket] = access
    return buckets_access


async def aio_s3_bucket_head(
    bucket_name: str, config: AioAWSConfig, s3_client: AioBaseClient
) -> Dict:
    """
    Asynchronous coroutine to issue a HEAD request for s3 bucket.

    :param bucket_name: an s3 bucket name
    :param config: an AioAWSConfig
    :param s3_client: an AioBaseClient for s3
    :return: a response to a HEAD request
    :raises: botocore.exceptions.ClientError
    """
    for tries in range(config.retries + 1):
        try:
            LOGGER.debug("AWS S3 bucket HEAD: %s", bucket_name)
            response = await s3_client.head_bucket(Bucket=bucket_name)
            LOGGER.debug("AWS S3 bucket HEAD: %s", response)
            return response

        except ClientError as err:
            LOGGER.debug("AWS S3 bucket HEAD error: %s", err.response)
            err_code = err.response.get("Error", {}).get("Code")
            if err_code == "TooManyRequestsException":
                if tries < config.retries:
                    await jitter("s3-bucket-head", config.min_jitter, config.max_jitter)
            else:
                raise

    raise RuntimeError("AWS S3 bucket HEAD exceeded retries")


async def aio_s3_bucket_access(
    bucket_name: str, config: AioAWSConfig, s3_client: AioBaseClient
) -> Tuple[str, bool]:
    """
    Asynchronous coroutine to issue a HEAD request to check if s3 bucket access is allowed.

    :param bucket_name: an s3 bucket name
    :param config: an AioAWSConfig
    :param s3_client: an AioBaseClient for s3
    :return: a tuple of ``(bucket_name: str, access: bool)``
    """
    try:
        head = await aio_s3_bucket_head(bucket_name, config=config, s3_client=s3_client)
        if response_success(head):
            return bucket_name, True
        return bucket_name, False
    except ClientError as err:
        access = handle_head_error_code(err, bucket_name)
        if access is False:
            return bucket_name, False
        raise


async def aio_s3_buckets_list(config: AioAWSConfig, s3_client: AioBaseClient) -> Dict:
    """
    Asynchronous coroutine to list all buckets.

    :param config: an AioAWSConfig
    :param s3_client: an AioBaseClient for s3
    :return: a response to a ListBuckets request
    """
    for tries in range(config.retries + 1):
        try:
            LOGGER.debug("AWS S3 buckets list")
            response = await s3_client.list_buckets()
            LOGGER.debug("AWS S3 buckets list: %s", response)
            return response

        except ClientError as err:
            LOGGER.debug("AWS S3 buckets list error: %s", err.response)
            err_code = err.response.get("Error", {}).get("Code")
            if err_code == "TooManyRequestsException":
                if tries < config.retries:
                    await jitter("s3-bucket-list", config.min_jitter, config.max_jitter)
            else:
                raise

    raise RuntimeError("AWS S3 buckets list exceeded retries")


async def aio_s3_buckets_access(
    buckets: List[str],
    config: AioAWSConfig,
    s3_client: AioBaseClient,
) -> Dict[str, bool]:
    """
    Asynchronous coroutine to issue HEAD requests on all available
    s3 buckets to check if each bucket allows access.

    :param buckets: a list of bucket names to check; the default is
                    to issue a request for all buckets in the session region
    :param config: an AioAWSConfig
    :param s3_client: an AioBaseClient for s3
    :return: dict of ``{bucket_name: str, access: bool}``
    """
    access_coroutines = [
        aio_s3_bucket_access(bucket, config=config, s3_client=s3_client)
        for bucket in buckets
    ]
    bucket_access = {}
    for task in asyncio.as_completed(access_coroutines):
        result = await task
        bucket, access = result
        bucket_access[bucket] = access

    return bucket_access


async def aio_s3_object_head(
    s3_uri: str, config: AioAWSConfig, s3_client: AioBaseClient
) -> Dict:
    """
    Asynchronous coroutine to issue a HEAD request for s3 object.

    :param s3_uri: an s3 URI
    :param config: an AioAWSConfig
    :param s3_client: an AioBaseClient for s3
    :return: a response to a HEAD request
    :raises: botocore.exceptions.ClientError
    """
    s3_parts = S3Parts.parse_s3_uri(s3_uri)
    for tries in range(config.retries + 1):
        try:
            LOGGER.debug("AWS S3 object HEAD: %s", s3_uri)
            response = await s3_client.head_object(
                Bucket=s3_parts.bucket, Key=s3_parts.key
            )
            LOGGER.debug("AWS S3 object HEAD: %s", response)
            return response

        except ClientError as err:
            LOGGER.debug("AWS S3 object HEAD error: %s", err.response)
            err_code = err.response.get("Error", {}).get("Code")
            if err_code == "TooManyRequestsException":
                if tries < config.retries:
                    await jitter("s3-object-head", config.min_jitter, config.max_jitter)
                continue  # allow it to retry, if possible
            else:
                raise

    raise RuntimeError("AWS S3 object HEAD exceeded retries")


async def aio_s3_object_access(
    s3_uri: str, config: AioAWSConfig, s3_client: AioBaseClient
) -> Tuple[str, bool]:
    """
    Asynchronous coroutine to issue a HEAD request to check if s3 object access is allowed.

    :param s3_uri: an s3 URI
    :param config: an AioAWSConfig
    :param s3_client: an AioBaseClient for s3
    :return: a tuple of ``(s3_uri: str, access: bool)``
    """
    try:
        head = await aio_s3_object_head(s3_uri, config=config, s3_client=s3_client)
        if response_success(head):
            return s3_uri, True
        return s3_uri, False
    except ClientError as err:
        access = handle_head_error_code(err, s3_uri)
        if access is False:
            return s3_uri, False
        raise


async def aio_s3_objects_list(
    bucket_name: str,
    bucket_prefix: str,
    config: AioAWSConfig,
    s3_client: AioBaseClient,
) -> List[Dict]:
    """
    Asynchronous coroutine to collect all objects in a bucket prefix.

    :param bucket_name: param passed to s3_client.list_objects_v2 'Bucket'
    :param bucket_prefix: param passed to s3_client.list_objects_v2 'Prefix'
    :param config: an AioAWSConfig
    :param s3_client: an AioBaseClient for s3
    :return: a list of s3 object data, e.g.

        .. code-block::

            >>> aio_s3_objects[0]
            {'ETag': '"192e29f360ea8297b5876b33b8419741"',
             'Key': 'ABI-L2-ADPC/2019/337/13/OR_ABI-L2-ADPC-M6_G16_s20193371331167_e20193371333539_c20193371334564.nc',
             'LastModified': datetime.datetime(2019, 12, 3, 14, 27, 5, tzinfo=tzutc()),
             'Size': 892913,
             'StorageClass': 'INTELLIGENT_TIERING'}

    .. seealso:
        - https://botocore.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.list_objects_v2
    """
    for tries in range(config.retries + 1):
        try:
            LOGGER.debug(
                "AWS S3 list objects, get first page: %s/%s",
                bucket_name,
                bucket_prefix,
            )
            response = await s3_client.list_objects_v2(
                Bucket=bucket_name, Prefix=bucket_prefix
            )
            LOGGER.debug("AWS S3 list objects OK? %s", response_success(response))

            s3_objects = []
            while True:
                if response_success(response):
                    s3_objects += response["Contents"]

                    if response["IsTruncated"]:
                        LOGGER.debug(
                            "AWS S3 list objects, get next page: %s/%s",
                            bucket_name,
                            bucket_prefix,
                        )
                        next_token = response["NextContinuationToken"]
                        response = await s3_client.list_objects_v2(
                            Bucket=bucket_name,
                            Prefix=bucket_prefix,
                            ContinuationToken=next_token,
                        )
                        LOGGER.debug(
                            "AWS S3 list objects OK? %s", response_success(response)
                        )
                        continue
                    else:
                        break
                else:
                    break

            return s3_objects

        except ClientError as err:
            LOGGER.debug("AWS S3 list objects error: %s", err.response)
            err_code = err.response.get("Error", {}).get("Code")
            if err_code == "TooManyRequestsException":
                if tries < config.retries:
                    await jitter(
                        "s3-list-objects", config.min_jitter, config.max_jitter
                    )
                continue  # allow it to retry, if possible
            else:
                raise

    raise RuntimeError("AWS S3 list objects exceeded retries")


async def aio_s3_objects_access(
    s3_uris: List[str],
    config: AioAWSConfig,
    s3_client: AioBaseClient,
) -> Dict[str, bool]:
    """
    Asynchronous coroutine to issue HEAD requests on all available
    s3 objects to check if each s3_object allows access.

    :param s3_uris: a list of s3 uris in the form 's3://bucket-name/key'
    :param config: an AioAWSConfig
    :param s3_client: an AioBaseClient for s3
    :return: dict of ``{s3_uri: str, access: bool}``
    """
    access_coroutines = [
        aio_s3_object_access(s3_uri, config=config, s3_client=s3_client)
        for s3_uri in s3_uris
    ]
    object_access = {}
    for task in asyncio.as_completed(access_coroutines):
        result = await task
        s3_uri, access = result
        object_access[s3_uri] = access

    return object_access


if __name__ == "__main__":

    # pylint: disable=C0103

    #
    # Integration tests against NOAA public data on S3
    #

    def summary_accessible_buckets(buckets):
        accessible_buckets = [k for k, v in buckets.items() if v]
        print(f"queried {len(buckets)} buckets")
        print(f"found {len(accessible_buckets)} accessible buckets")

    # https://registry.opendata.aws/noaa-goes/
    noaa_goes_bucket = "noaa-goes16"
    noaa_prefix = "ABI-L2-ADPC/2019"  # use a prior year for stable results

    async def aio_main():

        # Use a single session with a larger connection pool.
        # The AioAWSConfig can provide clients on-demand with
        # a context manager pattern.
        aio_session = aio_aws_session()
        aio_config = AioAWSConfig(
            max_pool_connections=40,
            session=aio_session,
            min_pause=0.2,
            max_pause=0.6,
            min_jitter=0.2,
            max_jitter=0.8,
        )

        client_config = aio_config.session.get_default_client_config()
        s3_config = AioConfig(signature_version=UNSIGNED)
        s3_config = client_config.merge(s3_config)

        async with aio_config.create_client("s3", config=s3_config) as s3_client:

            print()
            print("aio-aws check an s3 object that exists")
            s3_uri = f"s3://{noaa_goes_bucket}/index.html"
            head_response = await aio_s3_object_head(
                s3_uri, config=aio_config, s3_client=s3_client
            )
            print(f"GET HEAD {s3_uri} -> {head_response}\n")

            print()
            print("aio-aws check an s3 object that is missing; it raises.")
            try:
                s3_uri = f"s3://{noaa_goes_bucket}/missing.html"
                await aio_s3_object_head(s3_uri, config=aio_config, s3_client=s3_client)
            except ClientError as err:
                err_code = err.response.get("Error", {}).get("Code")
                if err_code == "401":
                    print(f"GET HEAD {s3_uri} -> 401: invalid credentials\n")
                if err_code == "403":
                    print(f"GET HEAD {s3_uri} -> 403: permission denied\n")
                if err_code == "404":
                    print(f"GET HEAD {s3_uri} -> 404: object missing\n")

            # # This fails with a strange HTML response and a boto3 error
            # # botocore.parsers.ResponseParserError - so maybe any
            # # UNSIGNED requests are not able to list-buckets?
            # print()
            # print("aio-aws find all buckets that allow access.")
            # start = time.perf_counter()
            # buckets_response = await aio_s3_buckets_list(
            #     config=aio_config, s3_client=s3_client
            # )
            # s3_buckets = [bucket["Name"] for bucket in buckets_response["Buckets"]]
            # aio_bucket_access = await aio_s3_buckets_access(
            #     s3_buckets, config=aio_config, s3_client=s3_client
            # )
            # end = time.perf_counter() - start
            # summary_accessible_buckets(aio_bucket_access)
            # print(f"finished in {end:0.2f} seconds.\n")

            print()
            print("aio-aws collection of all objects in a bucket-prefix.")
            start = time.perf_counter()
            aio_s3_objects = await aio_s3_objects_list(
                bucket_name=noaa_goes_bucket,
                bucket_prefix=noaa_prefix,
                config=aio_config,
                s3_client=s3_client,
            )
            aio_s3_uris = [
                f"s3://{noaa_goes_bucket}/{obj['Key']}" for obj in aio_s3_objects
            ]
            print(f"found {len(aio_s3_uris)} s3 objects")
            end = time.perf_counter() - start
            print(f"finished in {end:0.2f} seconds.\n")
            set_aio_s3_uris = set(aio_s3_uris)
            assert len(aio_s3_uris) == len(set_aio_s3_uris)  # are they unique?

            print()
            print("aio-aws checks for access to all objects in a bucket-prefix.")
            start = time.perf_counter()
            aio_s3_access = await aio_s3_objects_access(
                s3_uris=aio_s3_uris, config=aio_config, s3_client=s3_client
            )
            print(f"checked access to {len(aio_s3_access)} s3 objects")
            end = time.perf_counter() - start
            print(f"finished in {end:0.2f} seconds.\n")

    asyncio.run(aio_main())

    # run_seq = True
    # if run_seq:
    #     print()
    #     print("aws-sequential find all buckets that allow access.")
    #     start = time.perf_counter()
    #     boto_session = botocore.session.get_session()
    #     boto_client = boto_session.create_client("s3")
    #     boto_bucket_access = aws_s3_buckets_access(boto_client)
    #     end = time.perf_counter() - start
    #     summary_accessible_buckets(boto_bucket_access)
    #     print(f"finished in {end:0.2f} seconds.\n")
    #     assert boto_bucket_access == aio_bucket_access
    #
    #     print()
    #     print("aws-sequential collection for all objects in a bucket-prefix")
    #     start = time.perf_counter()
    #     s3_bucket = boto3.resource("s3").Bucket(noaa_goes_bucket)
    #     objects_collection = s3_bucket.objects.filter(Prefix=noaa_prefix)
    #     s3_uris = [
    #         f"s3://{s3_summary.bucket_name}/{s3_summary.key}"
    #         for s3_summary in objects_collection
    #     ]
    #     print(f"found {len(s3_uris)} s3 objects")
    #     end = time.perf_counter() - start
    #     print(f"finished in {end:0.2f} seconds.\n")
    #
    #     print()
    #     print("Checking equality of collections for aio-aws vs. aws-sync")
    #     set_s3_uris = set(s3_uris)
    #     assert len(s3_uris) == len(set_s3_uris)  # are they unique?
    #
    #     # same set of URIS?  Sometimes it's off-by-one or a few and
    #     # maybe that's due to s3-consistency and object additions or removal
    #     assert len(s3_uris) == len(aio_s3_uris)
    #     assert set_aio_s3_uris == set_s3_uris
    #     print("The collections for aio-aws vs. aws-sync match OK")



# Copyright 2020 Darren Weber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
AioAWS Settings
---------------

Using asyncio for AWS services requires the `aiobotocore`_ library, which wraps a
release of `botocore`_ to patch it with features for async coroutines using
`asyncio`_ and `aiohttp`_.  To avoid issuing too many concurrent requests (DOS attack),
the async approach should use a client connection limiter, based on ``asyncio.Semaphore()``.
It's recommended to use a single session and a single client with a connection pool.
Although there are context manager patterns, it's also possible to manage closing the client
after everything is done.

.. seealso::
    - https://aiobotocore.readthedocs.io/en/latest/
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html
    - https://www.mathewmarcus.com/blog/asynchronous-aws-api-requests-with-asyncio.html

.. _aioboto3: https://github.com/terrycain/aioboto3
.. _aiobotocore: https://github.com/aio-libs/aiobotocore
.. _aiohttp: https://aiohttp.readthedocs.io/en/latest/
.. _asyncio: https://docs.python.org/3/library/asyncio.html
.. _botocore: https://botocore.amazonaws.com/v1/documentation/api/latest/index.html
.. _TinyDB: https://tinydb.readthedocs.io/en/latest/intro.html
"""


import asyncio
import random
from contextlib import asynccontextmanager
from dataclasses import dataclass

import aiobotocore.client
import aiobotocore.config
import aiobotocore.session
import botocore.client
import botocore.endpoint

from aio_aws.logger import get_logger

LOGGER = get_logger(__name__)

#: max_pool_connections for AWS clients (10 by default)
MAX_POOL_CONNECTIONS = botocore.endpoint.MAX_POOL_CONNECTIONS

#: Minimum task pause
MIN_PAUSE: float = 5

#: Maximum task pause
MAX_PAUSE: float = 30

#: Minimum API request jitter
MIN_JITTER: float = 1

#: Maximum API request jitter
MAX_JITTER: float = 10


def asyncio_default_semaphore() -> asyncio.Semaphore:
    """
    a default semaphore to limit creation of clients;
    it defaults to 2 * py:const:`MAX_POOL_CONNECTIONS`

    .. seealso::
        https://github.com/boto/botocore/blob/develop/botocore/config.py

    :return: aiobotocore.config.AioConfig
    """
    return asyncio.Semaphore(MAX_POOL_CONNECTIONS * 2)


def aio_aws_default_config() -> aiobotocore.config.AioConfig:
    """
    Get a default asyncio AWS config using a default
    py:const:`MAX_POOL_CONNECTIONS`

    .. seealso::
        https://github.com/boto/botocore/blob/develop/botocore/config.py

    :return: aiobotocore.config.AioConfig
    """
    return aiobotocore.config.AioConfig(max_pool_connections=MAX_POOL_CONNECTIONS)


def aio_aws_default_session() -> aiobotocore.session.AioSession:
    """
    Get a default asyncio AWS session with a default config from
    :py:func:`aio_aws_default_config`

    :return: aiobotocore.session.AioSession
    """
    aio_config = aio_aws_default_config()
    aio_session = aiobotocore.get_session()
    aio_session.user_agent_name = "aio-aws"
    aio_session.set_default_client_config(aio_config)
    # aio_session.full_config
    # aio_session.get_config_variable('region')
    # aio_session.get_scoped_config() is the same as:
    #     configs = aio_session.full_config
    #     configs['profiles'][aio_session.profile]
    return aio_session


def aio_aws_session(
    aio_aws_config: aiobotocore.config.AioConfig = None,
) -> aiobotocore.session.AioSession:
    """
    Get an asyncio AWS session with an 'aio-aws' user agent name

    :param aio_aws_config: an aiobotocore.config.AioConfig
            (default :py:func:`aio_aws_default_config`)
    :return: aiobotocore.session.AioSession
    """
    if aio_aws_config is None:
        aio_aws_config = aio_aws_default_config()
    session = aiobotocore.get_session()
    session.user_agent_name = "aio-aws"
    # session.set_stream_logger("aio-aws")  # for debugging
    session.set_default_client_config(aio_aws_config)
    return session


@asynccontextmanager
async def aio_aws_client(
    service_name: str, *args, **kwargs
) -> aiobotocore.client.AioBaseClient:
    """
    Yield an asyncio AWS client with an option to provide a client-specific config; this is a
    thin wrapper on ``aiobotocore.get_session().create_client()`` and the additional
    kwargs as passed through to ``session.create_client(**kwargs)``.

    It is possible to pass through additional args and kwargs
    including `config: aiobotocore.config.AioConfig`
    (the default is :py:func:`aio_aws_default_config`)

    .. code-block::

        s3_endpoint = "http://localhost:5555"
        client_config = botocore.client.Config(
            read_timeout=120,
            max_pool_connections=50,
        )
        async with aio_aws_client(
            "s3", endpoint_url=s3_endpoint, config=client_config
        ) as client:
            assert "aiobotocore.client.S3" in str(client)
            assert isinstance(client, aiobotocore.client.AioBaseClient)

    :param service_name: an AWS service for a client, like "s3", try
            :py:meth:`session.get_available_services()`
    :yield: aiobotocore.client.AioBaseClient

    .. seealso::
        - https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """
    session = aio_aws_session()
    async with session.create_client(service_name, *args, **kwargs) as client:
        yield client


# TODO: consider ways to auto-wrap client services?
#       - explore function decorators for:
#         - using a client semaphore
#         - to handle client exceptions for too-many-requests
# AIO_AWS_SESSION.get_available_services()  # iterate on these
# model = AIO_AWS_SESSION.get_service_model('batch')
# model.operation_names  # like 'DescribeJobs'
# op = model.operation_model('DescribeJobs')
# >>> op.input_shape
# <StructureShape(DescribeJobsRequest)>
# >>> op.output_shape
# <StructureShape(DescribeJobsResponse)>
# >>> op.input_shape.members
# OrderedDict([('jobs', <ListShape(StringList)>)])
# >>> op.input_shape.required_members
# ['jobs']


@dataclass
class AioAWSConfig:
    #: an optional AWS region name
    aws_region: str = None
    #: a number of retries for an AWS client request/response
    retries: int = 4
    #: an asyncio.sleep for ``random.uniform(min_pause, max_pause)``
    min_pause: float = MIN_PAUSE
    #: an asyncio.sleep for ``random.uniform(min_pause, max_pause)``
    max_pause: float = MAX_PAUSE
    #: an asyncio.sleep for ``random.uniform(min_jitter, max_jitter)``
    min_jitter: float = MIN_JITTER
    #: an asyncio.sleep for ``random.uniform(min_jitter, max_jitter)``
    max_jitter: float = MAX_JITTER
    #: defines a limit to the number of client connections
    max_pool_connections: int = MAX_POOL_CONNECTIONS
    # An asyncio.Semaphore to limit the number of concurrent clients;
    # this defaults to the session client connection pool
    sem: int = MAX_POOL_CONNECTIONS
    #: an aiobotocore.session.AioSession
    session: aiobotocore.session.AioSession = None

    @classmethod
    def get_default_config(cls):
        # TODO: explore memoized default config
        return cls()

    def __post_init__(self):

        # see also aiobotocore.endpoint.AioEndpointCreator.create_endpoint
        # for all the options that config details can provide for aiohttp session

        if self.session is None:
            self.session = aio_aws_default_session()

        default_config = self.default_client_config

        # self.session.get_default_client_config() is Optional[botocore.client.Config]
        client_config = self.session.get_default_client_config()
        if client_config:
            client_config = client_config.merge(default_config)
        else:
            client_config = default_config

        self.session.set_default_client_config(client_config)

        # Lazy init for an asyncio instance
        self._semaphore = None

    @property
    def default_client_config(self) -> botocore.client.Config:
        # botocore/config.py lists all the options
        config = botocore.client.Config(
            connect_timeout=20,
            read_timeout=120,
            max_pool_connections=self.max_pool_connections,
        )
        if self.aws_region:
            config.region_name = self.aws_region
        return config

    @property
    def semaphore(self) -> asyncio.Semaphore:
        """
        An asyncio.Semaphore to limit the number of concurrent clients;
        this defaults to the session client connection pool
        """
        if self._semaphore is None:
            self._semaphore = asyncio.Semaphore(self.sem)
        return self._semaphore

    @asynccontextmanager
    async def create_client(
        self, service: str, *args, **kwargs
    ) -> aiobotocore.client.AioBaseClient:
        """
        Create and yield a new client using the ``AioAWSConfig.session``;
        the clients are configured with a default limit on the size of
        the connection pool defined by ``AioAWSConfig.max_pool_connections``

        .. code-block::

            config = AioAWSConfig()
            async with config.create_client("s3") as client:
                response = await s3_client.head_bucket(Bucket=bucket_name)

        It is possible to pass through additional args and kwargs
        including `config=botocore.client.Config`.

        :yield: an aiobotocore.client.AioBaseClient for AWS service

        .. seealso::
            - https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
        """
        async with self.session.create_client(service, *args, **kwargs) as client:
            yield client


async def delay(
    task_id: str, min_pause: float = MIN_PAUSE, max_pause: float = MAX_PAUSE
) -> float:
    """
    Await a random pause between :py:const:`MIN_PAUSE` and :py:const:`MAX_PAUSE`

    :param task_id: the ID for the task awaiting this pause
    :param min_pause: defaults to :py:const:`MIN_PAUSE`
    :param max_pause: defaults to :py:const:`MAX_PAUSE`
    :return: random interval for pause
    """
    rand_pause = random.uniform(min_pause, max_pause)
    LOGGER.debug("Task %s - await a sleep for %.2f", task_id, rand_pause)
    try:
        await asyncio.sleep(rand_pause)
        LOGGER.debug("Task %s - done with sleep for %.2f", task_id, rand_pause)
        return rand_pause

    except asyncio.CancelledError:
        LOGGER.error("Task %s - cancelled", task_id)
        raise


async def jitter(
    task_id: str = "jitter",
    min_jitter: float = MIN_JITTER,
    max_jitter: float = MAX_JITTER,
) -> float:
    """
    Await a random pause between `min_jitter` and `max_jitter`

    :param task_id: an optional ID for the task awaiting this jitter
    :param min_jitter: defaults to :py:const:`MIN_JITTER`
    :param max_jitter: defaults to :py:const:`MAX_JITTER`
    :return: random interval for pause
    """
    jit = await delay(task_id, min_jitter, max_jitter)
    return jit
